# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import torch
import dexsim
import numpy as np
from functools import cached_property

from dataclasses import dataclass
from typing import List, Sequence, Union

from dexsim.models import MeshObject
from dexsim.engine import PhysicsScene
from dexsim.types import SoftBodyGPUAPIReadWriteType
from embodichain.lab.sim.common import (
    BatchEntity,
)
from embodichain.utils.math import (
    matrix_from_euler,
)
from embodichain.utils import logger
from embodichain.lab.sim.cfg import (
    SoftObjectCfg,
)
from embodichain.utils.math import xyz_quat_to_4x4_matrix


@dataclass
class SoftBodyData:
    """Data manager for soft body

    Note:
        1. The pose data managed by dexsim is in the format of (qx, qy, qz, qw, x, y, z), but in EmbodiChain, we use (x, y, z, qw, qx, qy, qz) format.
    """

    def __init__(
        self, entities: List[MeshObject], ps: PhysicsScene, device: torch.device
    ) -> None:
        """Initialize the SoftBodyData.

        Args:
            entities (List[MeshObject]): List of MeshObjects representing the soft bodies.
            ps (PhysicsScene): The physics scene.
            device (torch.device): The device to use for the soft body data.
        """
        self.entities = entities
        # TODO: soft body data can only be stored in cuda device for now.
        self.device = device
        # TODO: inorder to retrieve arena position, we need to access the node of each entity.
        self._arena_positions = self._get_arena_position()
        self.ps = ps
        self.num_instances = len(entities)

        self.softbodies = [
            self.entities[i].get_physical_body() for i in range(self.num_instances)
        ]
        self.n_collision_vertices = self.softbodies[0].get_num_vertices()
        self.n_sim_vertices = self.softbodies[0].get_num_sim_vertices()

        self._rest_position_buffer = torch.empty(
            (self.num_instances, self.n_collision_vertices, 4),
            device=self.device,
            dtype=torch.float32,
        )
        for i, softbody in enumerate(self.softbodies):
            self._rest_position_buffer[i] = softbody.get_position_inv_mass_buffer()

        self._rest_sim_position_buffer = torch.empty(
            (self.num_instances, self.n_sim_vertices, 4),
            device=self.device,
            dtype=torch.float32,
        )

        for i, softbody in enumerate(self.softbodies):
            self._rest_sim_position_buffer[i] = (
                softbody.get_sim_position_inv_mass_buffer()
            )

        self._collision_position = torch.zeros(
            (self.num_instances, self.n_collision_vertices, 3),
            device=self.device,
            dtype=torch.float32,
        )
        self._sim_vertex_velocity = torch.zeros(
            (self.num_instances, self.n_sim_vertices, 3),
            device=self.device,
            dtype=torch.float32,
        )
        self._sim_vertex_position = torch.zeros(
            (self.num_instances, self.n_sim_vertices, 3),
            device=self.device,
            dtype=torch.float32,
        )

    def _get_arena_position(self):
        n_env = len(self.entities)
        arena_positions = torch.empty(
            (n_env, 3), device=self.device, dtype=torch.float32
        )
        for i, entity in enumerate(self.entities):
            arena = entity.node.get_parent()
            arena_position = arena.get_world_pose()[:3, 3]
            arena_positions[i] = torch.as_tensor(
                arena_position, device=self.device, dtype=torch.float32
            )
        return arena_positions

    @property
    def rest_collision_vertices(self):
        """Get the rest position buffer of the soft bodies."""
        return self._rest_position_buffer[:, :, :3].clone()

    @property
    def rest_sim_vertices(self):
        """Get the rest sim position buffer of the soft bodies."""
        return self._rest_sim_position_buffer[:, :, :3].clone()

    @property
    def collision_position(self):
        """Get the current vertex position buffer of the soft bodies."""
        for i, softbody in enumerate(self.soft_bodies):
            self._collision_position[i] = softbody.get_position_inv_mass_buffer()[:, :3]
        return self._collision_position.clone()

    @property
    def sim_vertex_position(self):
        """Get the current sim vertex position buffer of the soft bodies."""
        for i, softbody in enumerate(self.soft_bodies):
            self._sim_vertex_position[i] = softbody.get_sim_position_inv_mass_buffer()[
                :, :3
            ]
        return self._sim_vertex_position.clone()

    @property
    def sim_vertex_velocity(self):
        """Get the current vertex velocity buffer of the soft bodies."""
        for i, softbody in enumerate(self.soft_bodies):
            self._sim_vertex_velocity[i] = softbody.get_sim_position_inv_mass_buffer()[
                :, :3
            ]
        return self._sim_vertex_velocity.clone()


class SoftObject(BatchEntity):
    """SoftObject represents a batch of rigid body in the simulation."""

    def __init__(
        self,
        cfg: SoftObjectCfg,
        entities: List[MeshObject] = None,
        device: torch.device = torch.device("cpu"),
    ) -> None:
        self._world = dexsim.default_world()
        self._ps = self._world.get_physics_scene()
        self._all_indices = torch.arange(len(entities), dtype=torch.int32).tolist()

        self._data = SoftBodyData(entities=entities, ps=self._ps, device=device)

        self._world.update(0.001)

        super().__init__(cfg=cfg, entities=entities, device=device)

        # set default collision filter
        self._set_default_collision_filter()

    def _set_default_collision_filter(self) -> None:
        collision_filter_data = torch.zeros(
            size=(self.num_instances, 4), dtype=torch.int32
        )
        for i in range(self.num_instances):
            collision_filter_data[i, 0] = i
            collision_filter_data[i, 1] = 1
        self.set_collision_filter(collision_filter_data)

    def set_collision_filter(
        self, filter_data: torch.Tensor, env_ids: Sequence[int] | None = None
    ) -> None:
        """Set collision filter data for the rigid object.

        Args:
            filter_data (torch.Tensor): [N, 4] of int.
                First element of each object is arena id.
                If 2nd element is 0, the object will collision with all other objects in world.
                3rd and 4th elements are not used currently.

            env_ids (Sequence[int] | None): Environment indices. If None, then all indices are used.
        """
        local_env_ids = self._all_indices if env_ids is None else env_ids

        if len(local_env_ids) != len(filter_data):
            logger.log_error(
                f"Length of env_ids {len(local_env_ids)} does not match pose length {len(filter_data)}."
            )

        filter_data_np = filter_data.cpu().numpy().astype(np.uint32)
        for i, env_idx in enumerate(local_env_ids):
            self._entities[env_idx].get_physical_body().set_collision_filter_data(
                filter_data_np[i]
            )

    @property
    def body_data(self) -> SoftBodyData | None:
        """Get the soft body data manager for this rigid object.

        Returns:
            SoftBodyData | None: The rigid body data manager.
        """
        return self._data

    def set_local_pose(
        self, pose: torch.Tensor, env_ids: Sequence[int] | None = None
    ) -> None:
        """Set local pose of the rigid object.

        Args:
            pose (torch.Tensor): The local pose of the rigid object with shape (N, 7) or (N, 4, 4).
            env_ids (Sequence[int] | None): Environment indices. If None, then all indices are used.
        """
        local_env_ids = self._all_indices if env_ids is None else env_ids

        if len(local_env_ids) != len(pose):
            logger.log_error(
                f"Length of env_ids {len(local_env_ids)} does not match pose length {len(pose)}."
            )

        if pose.dim() == 2 and pose.shape[1] == 7:
            pose4x4 = xyz_quat_to_4x4_matrix(pose)
        elif pose.dim() == 3 and pose.shape[1:3] == (4, 4):
            pose4x4 = pose
        else:
            logger.log_error(
                f"Invalid pose shape {pose.shape}. Expected (N, 7) or (N, 4, 4)."
            )

        for i, env_idx in enumerate(local_env_ids):
            # TODO: soft body cannot directly set by `set_local_pose` currently.
            rest_collision_vertices = self.body_data.rest_collision_vertices[i]
            rest_sim_vertices = self.body_data.rest_sim_vertices[i]
            rotation = pose4x4[i][:3, :3]
            translation = pose4x4[i][:3, 3]

            # apply transformation to local rest vertices and back
            rest_collision_vertices_local = (
                rest_collision_vertices - self._data._arena_positions[i]
            )
            transformed_collision_vertices = (
                rest_collision_vertices_local @ rotation.T + translation
            )
            transformed_collision_vertices = (
                transformed_collision_vertices + self._data._arena_positions[i]
            )

            rest_sim_vertices_local = rest_sim_vertices - self._data._arena_positions[i]
            transformed_sim_vertices = (
                rest_sim_vertices_local @ rotation.T + translation
            )
            transformed_sim_vertices = (
                transformed_sim_vertices + self._data._arena_positions[i]
            )

            # apply vertices to soft body
            soft_body = self._entities[env_idx].get_physical_body()
            collision_position_buffer = soft_body.get_position_inv_mass_buffer()
            sim_position_buffer = soft_body.get_sim_position_inv_mass_buffer()
            sim_velocity_buffer = soft_body.get_sim_velocity_buffer()

            collision_position_buffer[:, :3] = transformed_collision_vertices
            sim_position_buffer[:, :3] = transformed_sim_vertices
            sim_velocity_buffer[:, :3] = 0.0

            soft_body.mark_dirty(SoftBodyGPUAPIReadWriteType.ALL)
            # TODO: currently soft body has no wake up interface, use set_wake_counter and pass in a positive value to wake it up
            soft_body.set_wake_counter(0.4)

    def get_rest_collision_vertices(self) -> torch.Tensor:
        """Get the rest collision vertices of the soft object.

        Returns:
            torch.Tensor: The rest collision vertices with shape (N, num_collision_vertices, 3).
        """
        return self.body_data.rest_collision_vertices

    def get_rest_sim_vertices(self) -> torch.Tensor:
        """Get the rest sim vertices of the soft object.

        Returns:
            torch.Tensor: The rest sim vertices with shape (N, num_sim_vertices, 3).
        """
        return self.body_data.rest_sim_vertices

    def get_current_collision_vertices(self) -> torch.Tensor:
        """Get the current collision vertices of the soft object.

        Returns:
            torch.Tensor: The current collision vertices with shape (N, num_collision_vertices, 3).
        """
        return self.body_data.collision_position_buffer

    def get_current_sim_vertices(self) -> torch.Tensor:
        """Get the current sim vertices of the soft object.

        Returns:
            torch.Tensor: The current sim vertices with shape (N, num_sim_vertices, 3).
        """
        return self.body_data.sim_vertex_position_buffer

    def get_current_sim_vertex_velocities(self) -> torch.Tensor:
        """Get the current sim vertex velocities of the soft object.

        Returns:
            torch.Tensor: The current sim vertex velocities with shape (N, num_sim_vertices, 3).
        """
        return self.body_data.sim_vertex_velocity_buffer

    def get_local_pose(self, to_matrix: bool = False) -> torch.Tensor:
        """Get local pose of the rigid object.

        Args:
            to_matrix (bool, optional): If True, return the pose as a 4x4 matrix. If False, return as (x, y, z, qw, qx, qy, qz). Defaults to False.

        Returns:
            torch.Tensor: The local pose of the rigid object with shape (N, 7) or (N, 4, 4) depending on `to_matrix`.
        """
        raise NotImplementedError("Getting local pose for SoftObject is not supported.")

    def reset(self, env_ids: Sequence[int] | None = None) -> None:
        local_env_ids = self._all_indices if env_ids is None else env_ids
        num_instances = len(local_env_ids)

        # TODO: set attr for soft body after loading in physics scene.

        # rest soft body to init_pos
        pos = torch.as_tensor(
            self.cfg.init_pos, dtype=torch.float32, device=self.device
        )
        rot = (
            torch.as_tensor(self.cfg.init_rot, dtype=torch.float32, device=self.device)
            * torch.pi
            / 180.0
        )
        pos = pos.unsqueeze(0).repeat(num_instances, 1)
        rot = rot.unsqueeze(0).repeat(num_instances, 1)
        mat = matrix_from_euler(rot, "XYZ")
        pose = (
            torch.eye(4, dtype=torch.float32, device=self.device)
            .unsqueeze(0)
            .repeat(num_instances, 1, 1)
        )
        pose[:, :3, 3] = pos
        pose[:, :3, :3] = mat
        self.set_local_pose(pose, env_ids=local_env_ids)

    def destroy(self) -> None:
        # TODO: not tested yet
        env = self._world.get_env()
        arenas = env.get_all_arenas()
        if len(arenas) == 0:
            arenas = [env]
        for i, entity in enumerate(self._entities):
            arenas[i].remove_actor(entity)



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import os
import torch
import numpy as np
import pickle

from copy import deepcopy
from typing import Sequence
from scipy.spatial.transform import Rotation as R

from embodichain.lab.gym.envs import EmbodiedEnv, EmbodiedEnvCfg
from embodichain.lab.gym.utils.registration import register_env
from embodichain.data import get_data_path
from embodichain.utils import logger
from tqdm import tqdm


@register_env("ScoopIce-v1", max_episode_steps=600)
class ScoopIce(EmbodiedEnv):
    def __init__(self, cfg: EmbodiedEnvCfg = None, **kwargs):
        super().__init__(cfg, **kwargs)

        self.affordance_datas = {}

        # TODO: hardcode code, should be implemented as functor way.
        self.trajectory = pickle.load(
            open(
                get_data_path("ScoopIceNewEnv/pose_record_20250919_184544.pkl"),
                "rb",
            )
        )
        self.trajectory_sample_rate = 2

    def set_scoop_pose(self, xyzrxryrz):
        scoop = self.sim.get_rigid_object("scoop")
        pose = np.eye(4)
        pose[:3, 3] = xyzrxryrz[:3]
        pose[:3, :3] = R.from_euler("XYZ", xyzrxryrz[3:], degrees=True).as_matrix()
        n_env = self.sim.num_envs
        pose_t = torch.tensor(
            pose[None, :, :].repeat(n_env, axis=0),
            dtype=torch.float32,
            device=self.device,
        )
        scoop.set_local_pose(pose_t)

    def set_cup_pose(self, xyzrxryrz):
        cup = self.sim.get_rigid_object("paper_cup")
        pose = np.eye(4)
        pose[:3, 3] = xyzrxryrz[:3]
        pose[:3, :3] = R.from_euler("XYZ", xyzrxryrz[3:], degrees=True).as_matrix()
        n_env = self.sim.num_envs
        pose_t = torch.tensor(
            pose[None, :, :].repeat(n_env, axis=0),
            dtype=torch.float32,
            device=self.device,
        )
        cup.set_local_pose(pose_t)

    def add_xpos_offset(self, arm_qpos: np.ndarray, offset: np.ndarray, is_left: bool):
        """Add offset to arm qposes along end-effector x axis.

        Args:
            arm_qposes (np.ndarray): [waypoint_num, dof]
        """
        waypoint_num = arm_qpos.shape[0]
        dof = arm_qpos.shape[1]
        offset_t = torch.tensor(offset, dtype=torch.float32, device=self.device)
        control_part = "left_arm" if is_left else "right_arm"

        arm_qpos_batch = torch.tensor(
            arm_qpos[None, :, :], dtype=torch.float32, device=self.device
        )

        arm_xpos_batch = self.robot.compute_batch_fk(
            qpos=arm_qpos_batch, name=control_part, to_matrix=True
        )
        arm_xpos_batch[:, :, :3, 3] += offset_t
        ret, arm_qpos_offset_batch = self.robot.compute_batch_ik(
            pose=arm_xpos_batch,
            joint_seed=arm_qpos_batch,
            name=control_part,
        )
        return arm_qpos_offset_batch[0].to("cpu").numpy()

    def pack_qpos(self):
        left_arm_qpos = self.trajectory["left_arm"]  # [waypoint_num, dof]
        logger.log_info("Adding x and z offset to left arm trajectory...")
        left_arm_qpos = self.add_xpos_offset(
            arm_qpos=left_arm_qpos, offset=np.array([-0.018, 0.0, -0.01]), is_left=True
        )
        right_arm_qpos = self.trajectory["right_arm"]  # [waypoint_num, dof]
        # TODO: add z offset to right arm
        logger.log_info("Adding z offset to right arm trajectory...")
        right_arm_qpos = self.add_xpos_offset(
            arm_qpos=right_arm_qpos, offset=np.array([0.00, 0.0, 0.02]), is_left=False
        )
        left_eef_qpos = self.trajectory["left_eef"]  # [waypoint_num, hand_dof]
        right_eef_qpos = self.trajectory["right_eef"]
        torso_qpos = self.trajectory["torso"]
        # TODO: need head qpos.

        left_arm_qpos_expand = left_arm_qpos[None, :, :].repeat(self.num_envs, axis=0)
        right_arm_qpos_expand = right_arm_qpos[None, :, :].repeat(self.num_envs, axis=0)
        left_eef_qpos_expand = left_eef_qpos[None, :, :].repeat(self.num_envs, axis=0)
        right_eef_qpos_expand = right_eef_qpos[None, :, :].repeat(self.num_envs, axis=0)
        torso_qpos_expand = torso_qpos[None, :, :].repeat(self.num_envs, axis=0)
        all_qpos = np.concatenate(
            [
                left_arm_qpos_expand,
                right_arm_qpos_expand,
                left_eef_qpos_expand,
                right_eef_qpos_expand,
                torso_qpos_expand,
            ],
            axis=2,
        )
        return all_qpos

    def _initialize_episode(
        self, env_ids: Sequence[int] | None = None, **kwargs
    ) -> None:

        left_arm_ids = self.robot.get_joint_ids(name="left_arm")
        right_arm_ids = self.robot.get_joint_ids(name="right_arm")
        left_eef_ids = self.robot.get_joint_ids(name="left_eef")
        right_eef_ids = self.robot.get_joint_ids(name="right_eef")
        torso_ids = self.robot.get_joint_ids(name="torso")
        all_ids = np.hstack(
            [left_arm_ids, right_arm_ids, left_eef_ids, right_eef_ids, torso_ids]
        )

        # TODO: read xy random range from config
        xy_random_range = np.array([[-0.01, -0.01], [0.01, 0.01]])
        xy_random_offset = np.zeros(shape=(self.num_envs, 2))
        for arena_id in range(self.num_envs):
            xy_random_offset[arena_id] = np.random.uniform(
                low=xy_random_range[0], high=xy_random_range[1], size=(2,)
            )
        # TODO: apply warping to container pose

        all_qpos = self.pack_qpos()
        all_qpos_t = torch.tensor(all_qpos, dtype=torch.float32, device=self.device)

        # to initial qpos
        left_open_qpos = np.array([0.06, 1.5, 0.2, 0.2, 0.2, 0.2])
        left_close_qpos = np.array([0.13, 1.5, 0.5, 0.5, 0.5, 0.5])
        right_open_qpos = np.array([0.3, 1.5, 0.3, 0.3, 0.3, 0.3])
        right_close_qpos = np.array([0.6, 1.5, 0.7, 0.5, 0.7, 0.6])

        all_qpos_t[:, :, 14:20] = torch.tensor(
            left_close_qpos, dtype=torch.float32, device=self.device
        )
        all_qpos_t[:, :, 20:26] = torch.tensor(
            right_close_qpos, dtype=torch.float32, device=self.device
        )

        first_close_qpos = all_qpos_t[:, 0, :].to("cpu").numpy()
        first_open_qpos = deepcopy(first_close_qpos)

        # to first open pose
        first_open_qpos[:, 14:20] = left_open_qpos
        first_open_qpos[:, 20:26] = right_open_qpos
        self.robot.set_qpos(
            torch.tensor(first_open_qpos, dtype=torch.float32, device=self.device),
            joint_ids=all_ids,
        )
        self.sim.update(step=200)
        # save warp trajectory as demo action list
        waypoint_num = self.trajectory["left_arm"].shape[0]
        current_qpos = self.robot.get_qpos()
        self.demo_action_list = []
        for waypoint_idx in range(waypoint_num):
            action = current_qpos.clone()
            action[:, all_ids] = all_qpos_t[:, waypoint_idx, :]
            # TODO: sample in trajectory
            self.demo_action_list.append(action)

            # TODO: tricky implementation. Hold the first joint state for a while.
            if waypoint_idx == 0:
                for _ in range(20):
                    self.demo_action_list.append(action)

        self.sim.update(step=100)

        # apply events such as randomization for environments that need a reset
        if self.cfg.events:
            if "reset" in self.event_manager.available_modes:
                self.event_manager.apply(mode="reset", env_ids=env_ids)

    def create_demo_action_list(self, *args, **kwargs):
        logger.log_info(
            f"The original demo action list length: {len(self.demo_action_list)}"
        )
        logger.log_info(
            f"Downsample the demo action list by self.trajectory_sample_rate5 times."
        )
        return self.demo_action_list[:: self.trajectory_sample_rate]



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import traceback
import numpy as np
from pathlib import Path
from dataclasses import dataclass
import xml.etree.ElementTree as ET
from typing import Dict

from embodichain.toolkits.urdf_assembly.logging_utils import (
    URDFAssemblyLogger,
)
from embodichain.toolkits.urdf_assembly.mesh import URDFMeshManager


__all__ = ["ComponentRegistry", "URDFComponent", "URDFComponentManager"]


class ComponentRegistry:
    r"""Registry for storing and retrieving URDFComponent objects."""

    def __init__(self):
        self._components = {}

    def add(self, component_type: str, component_obj):
        self._components[component_type] = component_obj

    def get(self, component_type: str):
        return self._components.get(component_type)

    def all(self):
        return self._components

    def remove(self, component_type: str):
        if component_type in self._components:
            del self._components[component_type]


@dataclass
class URDFComponent:
    r"""Represents a URDF component with its configuration and transformation.

    This dataclass encapsulates all the information needed to process and integrate
    a URDF component into the robot assembly, including file path, attachment
    configuration, parameters, and optional spatial transformation.
    """

    urdf_path: str  # Path to the URDF file for this component
    default_attach_link: str = (
        "base_link"  # Default link name for attachment (usually the first link)
    )
    params: Dict = None  # Component-specific parameters (e.g., wheel_type for chassis)
    transform: np.ndarray | None = (
        None  # Optional 4x4 transformation matrix for positioning
    )

    def __post_init__(self):
        # Convert path to Path object for better path handling
        if isinstance(self.urdf_path, str):
            self.urdf_path = Path(self.urdf_path)

        # Validate transformation matrix dimensions and type
        if self.transform is not None:
            if not isinstance(self.transform, np.ndarray) or self.transform.shape != (
                4,
                4,
            ):
                raise ValueError("Transform must be a 4x4 numpy array")


class URDFComponentManager:
    """Responsible for loading, renaming, and processing meshes for a single component."""

    def __init__(self, mesh_manager: URDFMeshManager):
        self.mesh_manager = mesh_manager
        self.logger = URDFAssemblyLogger.get_logger("component_manager")

    def process_component(
        self,
        comp: str,
        prefix: str,
        comp_obj,
        name_mapping: dict,
        base_points: dict,
        links: list,
        joints: list,
    ):
        r"""Process a single URDF component by renaming elements and handling meshes.

        Args:
            comp (str): Component name (e.g., 'chassis', 'left_arm', 'hand').
            prefix (str): Prefix to add to component elements for uniqueness (e.g., 'left_').
                        None means no prefix will be applied.
            comp_obj: URDFComponent object containing the component's URDF path and parameters.
            name_mapping (dict): Dictionary mapping (component, original_name) tuples to new names.
                            Used for resolving cross-references between components.
            base_points (dict): Dictionary mapping component names to their base connection link names.
                            Used for establishing parent-child relationships.
            links (list): Global list to collect all processed link elements from all components.
            joints (list): Global list to collect all processed joint elements from all components.
        """

        try:
            urdf_root = ET.parse(comp_obj.urdf_path).getroot()

            # Safe way to get link and joint names, handling None values
            global_link_names = {
                link.get("name").lower()
                for link in links
                if link.get("name") is not None
            }
            global_joint_names = {
                joint.get("name").upper()
                for joint in joints
                if joint.get("name") is not None
            }

            first_link_flag = True
            joint_name_mapping = {}

            # Process links first
            for link in urdf_root.findall("link"):
                orig_name = link.get("name")
                if orig_name is None:
                    self.logger.warning(
                        f"Found link with no name in component {comp}, skipping"
                    )
                    continue

                # Generate unique name
                if prefix:
                    new_name = self._generate_unique_name(
                        orig_name, prefix, global_link_names
                    ).lower()
                else:
                    # For components without prefix, ensure names are unique
                    if orig_name.lower() in global_link_names:
                        new_name = f"{comp}_{orig_name}".lower()
                    else:
                        new_name = orig_name.lower()

                global_link_names.add(new_name)

                # Set first link as base point
                if first_link_flag:
                    base_points[comp] = new_name
                    first_link_flag = False

                # Update link name mapping and set link name to lowercase
                name_mapping[(comp, orig_name)] = new_name
                link.set("name", new_name)
                links.append(link)

                self._process_meshes(link, comp_obj.urdf_path, comp)

            # Process joints: Build mapping table AND process all at once
            joints_to_process = []

            # First collect all joints and build complete mapping
            for joint in urdf_root.findall("joint"):
                orig_joint_name = joint.get("name")
                if orig_joint_name is None:
                    continue

                new_joint_name = self._generate_unique_name(
                    orig_joint_name, prefix, global_joint_names
                ).upper()
                global_joint_names.add(new_joint_name)

                # Build the complete mapping table
                joint_name_mapping[orig_joint_name] = new_joint_name
                joints_to_process.append((joint, orig_joint_name, new_joint_name))

            self.logger.debug(f"Joint name mapping for [{comp}]: {joint_name_mapping}")

            # Now process all joints with complete mapping available
            for joint, orig_joint_name, new_joint_name in joints_to_process:
                # Set the new joint name
                joint.set("name", new_joint_name)

                # Update parent and child links to lowercase - with None checks
                parent_elem = joint.find("parent")
                child_elem = joint.find("child")

                if parent_elem is not None:
                    parent = parent_elem.get("link")
                    if parent is not None:
                        new_parent_name = name_mapping.get(
                            (comp, parent), parent
                        ).lower()
                        parent_elem.set("link", new_parent_name)
                    else:
                        self.logger.warning(
                            f"Found parent element with no link attribute in joint {orig_joint_name}"
                        )

                if child_elem is not None:
                    child = child_elem.get("link")
                    if child is not None:
                        new_child_name = name_mapping.get((comp, child), child).lower()
                        child_elem.set("link", new_child_name)
                    else:
                        self.logger.warning(
                            f"Found child element with no link attribute in joint {orig_joint_name}"
                        )

                # Process mimic joint references using the complete mapping table
                mimic_elem = joint.find("mimic")
                if mimic_elem is not None:
                    mimic_joint = mimic_elem.get("joint")
                    if mimic_joint is not None:
                        self.logger.info(
                            f"Processing mimic joint reference: ({mimic_joint}) in joint ({orig_joint_name})"
                        )
                        # Look up the corresponding new joint name in the mapping table
                        new_mimic_joint = joint_name_mapping.get(mimic_joint)
                        if new_mimic_joint:
                            # Update the mimic element to reference the renamed joint
                            mimic_elem.set("joint", new_mimic_joint)
                            self.logger.info(
                                f"✓ Updated mimic joint reference: ({mimic_joint}) -> ({new_mimic_joint})"
                            )
                        else:
                            self.logger.warning(
                                f"✗ Could not find mapping for mimic joint: ({mimic_joint})"
                            )
                            self.logger.warning(
                                f"Available mappings: {list(joint_name_mapping.keys())}"
                            )

                joints.append(joint)

            self.logger.debug(
                f"Processed component: [{comp}], links: {len(urdf_root.findall('link'))}, joints: {len(urdf_root.findall('joint'))}"
            )

        except Exception as e:
            self.logger.error(
                f"Failed to process component [{comp}]: {e}", exc_info=True
            )
            self.logger.error(f"Traceback: {traceback.format_exc()}")

    def _generate_unique_name(
        self, orig_name: str, prefix: str, existing_names: set
    ) -> str:
        r"""Generate a unique name by adding a prefix and ensuring no conflicts.

        Args:
            orig_name (str): The original name to modify.
            prefix (str): The prefix to add to the name.
            existing_names (set): A set of existing names to check for conflicts.

        Returns:
            str: A unique name derived from the original name.
        """
        if orig_name is None:
            orig_name = "unnamed"

        if prefix and not orig_name.lower().startswith(prefix.lower()):
            new_name = f"{prefix}{orig_name}".lower()
        else:
            new_name = orig_name.lower()

        # Ensure the new name is unique
        if new_name in existing_names:
            counter = 1
            unique_name = f"{new_name}_{counter}"
            while unique_name in existing_names:
                counter += 1
                unique_name = f"{new_name}_{counter}"
            new_name = unique_name

        return new_name

    def _process_meshes(self, link: ET.Element, base_urdf_path: str, comp_name: str):
        r"""Process visual and collision meshes for a link.

        Args:
            link (ET.Element): The URDF link element to process.
            base_urdf_path (str): The base path for the URDF files.
            comp_name (str): The name of the component being processed.
        """
        try:
            for visual in link.findall("visual"):
                geometry = visual.find("geometry")
                if geometry is not None:
                    mesh = geometry.find("mesh")
                    if mesh is not None:
                        filename = mesh.get("filename")
                        if filename is not None:
                            self.logger.debug(f"Processing visual mesh: {filename}")
                            new_mesh_filename = (
                                self.mesh_manager.copy_and_modify_mesh_file(
                                    base_urdf_path,
                                    filename,
                                    "Visual",
                                    comp_name,
                                )
                            )
                            self.logger.debug(
                                f"Updated visual mesh filename: {new_mesh_filename}"
                            )
                            mesh.set("filename", new_mesh_filename)

            for collision in link.findall("collision"):
                geometry = collision.find("geometry")
                if geometry is not None:
                    mesh = geometry.find("mesh")
                    if mesh is not None:
                        filename = mesh.get("filename")
                        if filename is not None:
                            self.logger.debug(f"Processing collision mesh: {filename}")
                            new_mesh_filename = (
                                self.mesh_manager.copy_and_modify_mesh_file(
                                    base_urdf_path,
                                    filename,
                                    "Collision",
                                    comp_name,
                                )
                            )
                            self.logger.debug(
                                f"Updated collision mesh filename: {new_mesh_filename}"
                            )
                            mesh.set("filename", new_mesh_filename)
        except Exception as e:
            self.logger.error(
                f"Failed to process meshes for component {comp_name}: {e}"
            )



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import torch

from typing import Union


def standardize_device_string(device: Union[str, torch.device]) -> str:
    """Standardize the device string for Warp compatibility.

    Args:
        device (Union[str, torch.device]): The device specification.

    Returns:
        str: The standardized device string.
    """
    if isinstance(device, str):
        device_str = device
    else:
        device_str = str(device)

    if device_str.startswith("cuda"):
        device_str = "cuda:0"

    return device_str



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
import time
import torch
import pytest
import numpy as np
from copy import deepcopy
from embodichain.lab.sim import SimulationManager, SimulationManagerCfg
from embodichain.lab.sim.objects import Robot
from embodichain.lab.sim.robots import CobotMagicCfg

from embodichain.lab.sim.planners.utils import TrajectorySampleMethod
from embodichain.lab.sim.planners import (
    MotionGenerator,
    MotionGenCfg,
    MotionGenOptions,
    ToppraPlannerCfg,
    ToppraPlanOptions,
    PlanState,
    MoveType,
    MovePart,
)


def to_numpy(tensor):
    if isinstance(tensor, torch.Tensor):
        tensor = tensor.detach().cpu()
        if tensor.ndim == 3 and tensor.shape[0] == 1:
            tensor = tensor[0]
        return tensor.numpy()
    return np.array(tensor)


class BaseTestMotionGenerator(object):
    @classmethod
    def setup_class(cls):
        cls.config = SimulationManagerCfg(headless=True, sim_device="cpu")
        cls.robot_sim = SimulationManager(cls.config)
        cls.robot_sim.set_manual_update(False)

        cfg_dict = {
            "uid": "CobotMagic",
            "init_pos": [0.0, 0.0, 0.7775],
            "init_qpos": [
                -0.3,
                0.3,
                1.0,
                1.0,
                -1.2,
                -1.2,
                0.0,
                0.0,
                0.6,
                0.6,
                0.0,
                0.0,
                0.05,
                0.05,
                0.05,
                0.05,
            ],
            "solver_cfg": {
                "left_arm": {
                    "class_type": "OPWSolver",
                    "end_link_name": "left_link6",
                    "root_link_name": "left_arm_base",
                    "tcp": [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0.143], [0, 0, 0, 1]],
                },
                "right_arm": {
                    "class_type": "OPWSolver",
                    "end_link_name": "right_link6",
                    "root_link_name": "right_arm_base",
                    "tcp": [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0.143], [0, 0, 0, 1]],
                },
            },
        }

        cls.robot: Robot = cls.robot_sim.add_robot(
            cfg=CobotMagicCfg.from_dict(cfg_dict)
        )

        cls.arm_name = "left_arm"

        cls.motion_cfg = MotionGenCfg(
            planner_cfg=ToppraPlannerCfg(
                robot_uid=cls.robot.uid,
            )
        )
        cls.motion_gen = MotionGenerator(cfg=cls.motion_cfg)

        # Test data for trajectory generation
        qpos_fk = torch.tensor(
            [[0.0, np.pi / 4, -np.pi / 4, 0.0, np.pi / 4, 0.0]], dtype=torch.float32
        )
        xpos_begin = cls.robot.compute_fk(
            name=cls.arm_name, qpos=qpos_fk, to_matrix=True
        )
        xpos_mid = deepcopy(xpos_begin)
        xpos_mid[0, 2, 3] -= 0.1  # Move down by 0.1m in Z direction
        xpos_final = deepcopy(xpos_mid)
        xpos_final[0, 0, 3] += 0.2  # Move forward by 0.2m in X direction

        qpos_begin = cls.robot.compute_ik(pose=xpos_begin, name=cls.arm_name)[1][0]
        qpos_mid = cls.robot.compute_ik(pose=xpos_mid, name=cls.arm_name)[1][0]
        qpos_final = cls.robot.compute_ik(pose=xpos_final, name=cls.arm_name)[1][0]

        cls.qpos_list = [qpos_begin, qpos_mid, qpos_final]
        cls.xpos_list = [
            xpos_begin[0],
            xpos_mid[0],
            xpos_final[0],
        ]

        cls.sample_num = 20

    def get_joint_ids(self):
        return self.robot.get_joint_ids(self.arm_name)

    def get_current_qpos(self):
        qpos_tensor = self.robot.get_qpos()
        if qpos_tensor.ndim == 2 and qpos_tensor.shape[0] == 1:
            qpos_tensor = qpos_tensor[0]
        return qpos_tensor[self.get_joint_ids()].cpu()

    def verify_final_xpos(self, expected_xpos, decimal=5e-3):
        final_xpos = self.robot.compute_fk(
            qpos=self.get_current_qpos(), name=self.arm_name, to_matrix=True
        )
        np.testing.assert_array_almost_equal(
            to_numpy(final_xpos)[:3, 3],
            to_numpy(expected_xpos)[:3, 3],
            decimal=decimal,
            err_msg=f"Expected: {to_numpy(expected_xpos)[:3, 3]}, Got: {to_numpy(final_xpos)[:3, 3]}",
        )

    def _execute_trajectory(self, qpos_list, forward=True, delay=0.01):
        indices = (
            range(len(qpos_list)) if forward else range(len(qpos_list) - 1, -1, -1)
        )
        for i in indices:
            self.robot.set_qpos(qpos=qpos_list[i], joint_ids=self.get_joint_ids())
            time.sleep(delay)
        time.sleep(delay * 2)

    @classmethod
    def teardown_class(cls):
        try:
            cls.robot_sim.destroy()
            print("robot_sim destroyed successfully")
        except Exception as e:
            print(f"Error during robot_sim.destroy(): {e}")

    def _execute_forward_trajectory(self, robot, qpos_list, delay=0.1):
        """Helper method to execute trajectory"""
        # Forward
        for q in qpos_list:
            robot.set_qpos(qpos=q, joint_ids=self.robot.get_joint_ids(self.arm_name))
            time.sleep(delay)
        time.sleep(delay * 5)

    def _execute_backward_trajectory(self, robot, qpos_list, delay=0.1):
        """Helper method to execute trajectory"""
        # Backward
        for q in qpos_list[::-1]:
            robot.set_qpos(qpos=q, joint_ids=self.robot.get_joint_ids(self.arm_name))
            time.sleep(delay)
        time.sleep(delay * 5)


class TestMotionGenerator(BaseTestMotionGenerator):
    """Test suite for MotionGenerator trajectory generation"""

    @pytest.mark.parametrize("is_linear", [True, False])
    def test_create_trajectory_with_xpos(self, is_linear):
        """Test trajectory generation with cartesian positions"""
        self.robot.set_qpos(qpos=self.qpos_list[0], joint_ids=self.get_joint_ids())
        time.sleep(0.2)

        options = MotionGenOptions(
            start_qpos=self.qpos_list[0],
            control_part=self.arm_name,
            is_linear=is_linear,
            is_interpolate=True,
            plan_opts=ToppraPlanOptions(
                sample_method=TrajectorySampleMethod.QUANTITY,
                sample_interval=20,
                constraints={
                    "velocity": 0.2,
                    "acceleration": 0.5,
                },
            ),
        )
        target_states = []
        for xpos in self.xpos_list:
            target_states.append(
                PlanState(
                    move_type=MoveType.EEF_MOVE,
                    move_part=MovePart.LEFT,
                    xpos=xpos,
                )
            )

        plan_result = self.motion_gen.generate(
            target_states=target_states, options=options
        )
        out_qpos_list = to_numpy(plan_result.positions)
        assert (
            len(out_qpos_list) == self.sample_num
        ), f"Sample number mismatch: {len(out_qpos_list)} != {self.sample_num}"
        test_xpos = self.robot.compute_fk(
            qpos=plan_result.positions[-1].unsqueeze(0),
            name=self.arm_name,
            to_matrix=True,
        )[0]
        np.testing.assert_array_almost_equal(test_xpos, self.xpos_list[-1], decimal=3)
        self._execute_trajectory(out_qpos_list, forward=True)
        self.verify_final_xpos(self.xpos_list[-1])
        self._execute_trajectory(out_qpos_list, forward=False)
        self.verify_final_xpos(self.xpos_list[0])

    @pytest.mark.parametrize("is_linear", [True, False])
    def test_create_trajectory_with_qpos(self, is_linear):
        """Test trajectory generation with joint positions"""
        self.robot.set_qpos(qpos=self.qpos_list[0], joint_ids=self.get_joint_ids())
        time.sleep(0.05)

        options = MotionGenOptions(
            start_qpos=self.qpos_list[0],
            control_part=self.arm_name,
            is_linear=is_linear,
            is_interpolate=True,
            plan_opts=ToppraPlanOptions(
                sample_method=TrajectorySampleMethod.QUANTITY,
                sample_interval=20,
                constraints={
                    "velocity": 0.2,
                    "acceleration": 0.5,
                },
            ),
        )

        target_states = []
        for qpos in self.qpos_list:
            target_states.append(
                PlanState(
                    move_type=MoveType.JOINT_MOVE,
                    move_part=MovePart.LEFT,
                    qpos=qpos,
                )
            )
        plan_result = self.motion_gen.generate(
            target_states=target_states, options=options
        )
        out_qpos_list = to_numpy(plan_result.positions)
        assert (
            len(out_qpos_list) == self.sample_num
        ), f"Sample number mismatch: {len(out_qpos_list)} != {self.sample_num}"
        np.testing.assert_array_almost_equal(
            out_qpos_list[-1], self.qpos_list[-1], decimal=3
        )
        self._execute_trajectory(out_qpos_list, forward=True)
        self.verify_final_xpos(self.xpos_list[-1])
        self._execute_trajectory(out_qpos_list, forward=False)
        self.verify_final_xpos(self.xpos_list[0])

    @pytest.mark.parametrize("xpos_or_qpos", ["xpos", "qpos"])
    def test_estimate_trajectory_sample_count(self, xpos_or_qpos: str):
        """Test estimation of trajectory sample count"""
        if xpos_or_qpos == "xpos":
            estimated_num = self.motion_gen.estimate_trajectory_sample_count(
                xpos_list=torch.as_tensor(np.array(self.xpos_list)),
                step_size=0.01,
                angle_step=np.pi / 90,
                control_part=self.arm_name,
            )
        else:
            estimated_num = self.motion_gen.estimate_trajectory_sample_count(
                qpos_list=torch.as_tensor(np.array(self.qpos_list)),
                step_size=0.01,
                angle_step=np.pi / 90,
                control_part=self.arm_name,
            )
        assert (estimated_num - 30) < 2, "Estimated sample count failed"


if __name__ == "__main__":
    np.set_printoptions(precision=5, suppress=True)
    torch.set_printoptions(precision=5, sci_mode=False)
    pytest_args = ["-v", "-s", __file__]
    pytest.main(pytest_args)



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import os
import open3d as o3d

from pathlib import Path
from typing import List

from embodichain.data.dataset import EmbodiChainDataset
from embodichain.utils import logger
from embodichain.data.constants import (
    EMBODICHAIN_DOWNLOAD_PREFIX,
    EMBODICHAIN_DEFAULT_DATA_ROOT,
)


material_assets = "materials"


class SimResources(EmbodiChainDataset):
    def __init__(self, data_root: str = None):
        data_descriptor = o3d.data.DataDescriptor(
            os.path.join(
                EMBODICHAIN_DOWNLOAD_PREFIX, material_assets, "embodisim_resources.zip"
            ),
            "53c054b3ae0857416dc52632eb562c12",
        )
        prefix = "SimResources"
        path = EMBODICHAIN_DEFAULT_DATA_ROOT if data_root is None else data_root

        super().__init__(prefix, data_descriptor, path)

    def get_ibl_path(self, name: str) -> str:
        """Get the path of the IBL resource.

        Args:
            name (str): The name of the IBL resource.

        Returns:
            str: The path to the IBL resource.
        """
        ibl_names = self.get_ibl_list()
        if name not in ibl_names:
            logger.log_error(
                f"Invalid IBL name: {name}. Available names are: {ibl_names}"
            )
        return str(Path(self.extract_dir) / "embodysim_resources" / "IBL" / name)

    def get_ibl_list(self) -> List[str]:
        """Get the names of all IBL resources.

        Returns:
            List[str]: The names of all IBL resources.
        """
        return [
            f.name
            for f in Path(self.extract_dir).glob("embodysim_resources/IBL/*")
            if f.is_dir()
        ]

    def get_material_path(self, name: str) -> str:
        """Get the path of the material resource.

        Args:
            name (str): The name of the material resource.

        Returns:
            str: The path to the material resource.
        """
        material_names = self.get_material_list()
        if name not in material_names:
            logger.log_error(
                f"Invalid material name: {name}. Available names are: {material_names}"
            )
        return str(Path(self.extract_dir) / "embodysim_resources" / "materials" / name)

    def get_material_list(self) -> List[str]:
        """Get the names of all material resources.

        Returns:
            List[str]: The names of all material resources.
        """
        return [
            f.name
            for f in Path(self.extract_dir).glob("embodysim_resources/materials/*")
            if f.is_dir()
        ]


class CocoBackground(EmbodiChainDataset):
    def __init__(self, data_root: str = None):
        data_descriptor = o3d.data.DataDescriptor(
            os.path.join(
                EMBODICHAIN_DOWNLOAD_PREFIX, material_assets, "CocoBackground.zip"
            ),
            "fda82404a317281263bd5849e9eb31a1",
        )
        prefix = "CocoBackground"
        path = EMBODICHAIN_DEFAULT_DATA_ROOT if data_root is None else data_root

        super().__init__(prefix, data_descriptor, path)



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

from .base_env import *
from .embodied_env import *
from .tasks import *
from .wrapper import *

from embodichain.lab.gym.envs.embodied_env import EmbodiedEnv

# Specific task environments
from embodichain.lab.gym.envs.tasks.tableware.pour_water.pour_water import (
    PourWaterEnv,
    PourWaterAgentEnv,
)
from embodichain.lab.gym.envs.tasks.tableware.scoop_ice import ScoopIce
from embodichain.lab.gym.envs.tasks.tableware.stack_blocks_two import StackBlocksTwoEnv
from embodichain.lab.gym.envs.tasks.tableware.blocks_ranking_rgb import (
    BlocksRankingRGBEnv,
)
from embodichain.lab.gym.envs.tasks.tableware.blocks_ranking_size import (
    BlocksRankingSizeEnv,
)
from embodichain.lab.gym.envs.tasks.tableware.place_object_drawer import (
    PlaceObjectDrawerEnv,
)
from embodichain.lab.gym.envs.tasks.tableware.stack_cups import (
    StackCupsEnv,
)
from embodichain.lab.gym.envs.tasks.tableware.match_object_container import (
    MatchObjectContainerEnv,
)
from embodichain.lab.gym.envs.tasks.tableware.rearrangement import (
    RearrangementEnv,
    RearrangementAgentEnv,
)

# Reinforcement learning environments
from embodichain.lab.gym.envs.tasks.rl.push_cube import PushCubeEnv

from embodichain.lab.gym.envs.tasks.rl.basic.cart_pole import CartPoleEnv

from embodichain.lab.gym.envs.tasks.special.simple_task import SimpleTaskEnv



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import os

from .constants import EMBODICHAIN_DEFAULT_DATABASE_ROOT

database_dir = EMBODICHAIN_DEFAULT_DATABASE_ROOT
database_2d_dir = os.path.join(database_dir, "2dasset")
database_agent_prompt_dir = os.path.join(database_dir, "agent_prompt")
database_demo_dir = os.path.join(database_dir, "demostration")

from . import assets
from .dataset import *



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

from .kernels import reshape_tiled_image
from . import kinematics
from .kinematics.opw_solver import opw_fk_kernel, opw_ik_kernel
from .kinematics.warp_trajectory import (
    trajectory_get_diff_kernel,
    trajectory_interpolate_kernel,
    trajectory_add_origin_kernel,
    get_offset_qpos_kernel,
)

from .kinematics.interpolate import (
    pairwise_distances,
    cumsum_distances,
    repeat_first_point,
    interpolate_along_distance,
)



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import torch
import numpy as np

from embodichain.lab.gym.envs import EmbodiedEnv, EmbodiedEnvCfg
from embodichain.lab.gym.utils.registration import register_env
from embodichain.utils import logger

__all__ = ["StackCupsEnv"]


@register_env("StackCups-v1", max_episode_steps=600)
class StackCupsEnv(EmbodiedEnv):
    def __init__(self, cfg: EmbodiedEnvCfg = None, **kwargs):
        super().__init__(cfg, **kwargs)

        action_config = kwargs.get("action_config", None)
        if action_config is not None:
            self.action_config = action_config

    def is_task_success(self, **kwargs) -> torch.Tensor:
        """Determine if the task is successfully completed.

        The task is successful if:
        1. Both cups haven't fallen over
        2. Cups are aligned in xy plane
        3. Top cup is only slightly higher than base cup

        Args:
            **kwargs: Additional arguments for task-specific success criteria.

        Returns:
            torch.Tensor: A boolean tensor indicating success for each environment in the batch.
        """
        try:
            cup1 = self.sim.get_rigid_object("cup_1")
            cup2 = self.sim.get_rigid_object("cup_2")
        except Exception as e:
            logger.log_warning(f"Cups not found: {e}, returning False.")
            return torch.zeros(self.num_envs, dtype=torch.bool, device=self.device)

        # Get cup poses
        cup1_pose = cup1.get_local_pose(to_matrix=True)
        cup2_pose = cup2.get_local_pose(to_matrix=True)

        # Extract positions
        cup1_pos = cup1_pose[:, :3, 3]  # (num_envs, 3)
        cup2_pos = cup2_pose[:, :3, 3]  # (num_envs, 3)

        # Check if cups haven't fallen
        cup1_fallen = self._is_fall(cup1_pose)
        cup2_fallen = self._is_fall(cup2_pose)

        # Determine which cup is lower (should be cup1)
        cup1_is_lower = cup1_pos[:, 2] < cup2_pos[:, 2]

        # Use the lower cup as base, calculate expected position of top cup
        base_cup_pos = torch.where(
            cup1_is_lower.unsqueeze(1).expand_as(cup1_pos), cup1_pos, cup2_pos
        )
        top_cup_pos = torch.where(
            cup1_is_lower.unsqueeze(1).expand_as(cup2_pos), cup2_pos, cup1_pos
        )

        # Top cup should be slightly higher than base cup
        cup_rim_offset = 0.015
        expected_top_pos = torch.stack(
            [
                base_cup_pos[:, 0],
                base_cup_pos[:, 1],
                base_cup_pos[:, 2] + cup_rim_offset,
            ],
            dim=1,
        )

        # Tolerance
        eps = torch.tensor([0.04, 0.04, 0.02], dtype=torch.float32, device=self.device)

        # Check if top cup is within tolerance of expected position
        position_diff = torch.abs(top_cup_pos - expected_top_pos)  # (num_envs, 3)
        stacked_correctly = torch.all(
            position_diff < eps.unsqueeze(0), dim=1
        )  # (num_envs,)

        # Task succeeds if cups are stacked correctly and haven't fallen
        success = stacked_correctly & ~cup1_fallen & ~cup2_fallen

        return success

    def _is_fall(self, pose: torch.Tensor) -> torch.Tensor:
        # Extract z-axis from rotation matrix (last column, first 3 elements)
        pose_rz = pose[:, :3, 2]
        world_z_axis = torch.tensor([0, 0, 1], dtype=pose.dtype, device=pose.device)

        # Compute dot product for each batch element
        dot_product = torch.sum(pose_rz * world_z_axis, dim=-1)  # Shape: (batch_size,)

        # Clamp to avoid numerical issues with arccos
        dot_product = torch.clamp(dot_product, -1.0, 1.0)

        # Compute angle and check if fallen
        angle = torch.arccos(dot_product)
        return angle >= torch.pi / 4



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

from __future__ import annotations

import torch
from typing import TYPE_CHECKING

from embodichain.lab.sim.objects import RigidObject, Robot
from embodichain.lab.gym.envs.managers.cfg import SceneEntityCfg
from embodichain.utils.math import sample_uniform
from embodichain.utils import logger


if TYPE_CHECKING:
    from embodichain.lab.gym.envs import EmbodiedEnv


def randomize_rigid_object_mass(
    env: EmbodiedEnv,
    env_ids: torch.Tensor | list[int],
    entity_cfg: SceneEntityCfg,
    mass_range: tuple[float, float],
    relative: bool = False,
) -> None:
    """Randomize the mass of rigid objects in the environment.

    Args:
        env (EmbodiedEnv): The environment instance.
        env_ids (torch.Tensor | list[int]): The environment IDs to apply the randomization.
        entity_cfg (SceneEntityCfg): The configuration for the scene entity.
        mass_range (tuple[float, float]): The range (min, max) to sample the mass from.
        relative (bool): Whether to apply the mass change relative to the initial mass. Defaults to False.
    """

    if entity_cfg.uid not in env.sim.get_rigid_object_uid_list():
        return

    rigid_object: RigidObject = env.sim.get_rigid_object(entity_cfg.uid)
    num_instance = len(env_ids)

    sampled_masses = sample_uniform(
        lower=mass_range[0], upper=mass_range[1], size=(num_instance,)
    )

    if relative:
        init_mass = rigid_object.cfg.attrs.mass
        init_mass = torch.full((sampled_masses.shape), init_mass, device=env.device)
        sampled_masses = init_mass + sampled_masses

    rigid_object.set_mass(sampled_masses, env_ids=env_ids)


def randomize_rigid_object_center_of_mass(
    env: EmbodiedEnv,
    env_ids: torch.Tensor | list[int],
    entity_cfg: SceneEntityCfg,
    com_pos_offset_range: tuple[list[float], list[float]],
) -> None:
    """Randomize the center of mass of rigid objects in the environment.

    Args:
        env (EmbodiedEnv): The environment instance.
        env_ids (torch.Tensor | list[int]): The environment IDs to apply the randomization.
        entity_cfg (SceneEntityCfg): The configuration for the scene entity.
        com_pos_offset_range (tuple[list[float], list[float]]): The range (min, max) to sample the center of mass offset from.
    """

    if entity_cfg.uid not in env.sim.get_rigid_object_uid_list():
        return

    rigid_object: RigidObject = env.sim.get_rigid_object(entity_cfg.uid)
    if rigid_object.is_non_dynamic:
        logger.log_warning(
            f"Cannot randomize center of mass for non-dynamic rigid object '{entity_cfg.uid}'."
        )
        return

    num_instance = len(env_ids)

    sampled_com_pos_offsets = sample_uniform(
        lower=com_pos_offset_range[0],
        upper=com_pos_offset_range[1],
        size=(num_instance, 3),
    )

    com = rigid_object.body_data.default_com_pose[env_ids]
    updated_com = com.clone()
    updated_com[:, 0:3] += sampled_com_pos_offsets

    rigid_object.set_com_pose(updated_com, env_ids=env_ids)



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""
Gizmo Scene Example: Interactive scene with both robot and rigid object gizmos

This example demonstrates how to create an interactive simulation scene with:
- A UR10 robot with gizmo control for end-effector manipulation
- A rigid object (cube) with gizmo control for direct manipulation
Both objects can be interactively controlled through their respective gizmos.
"""

import time
import torch
import numpy as np
import argparse
import cv2

from embodichain.lab.sim import SimulationManager, SimulationManagerCfg
from embodichain.lab.sim.cfg import (
    RobotCfg,
    URDFCfg,
    JointDrivePropertiesCfg,
    RigidObjectCfg,
    RigidBodyAttributesCfg,
)
from embodichain.lab.sim.shapes import CubeCfg
from embodichain.lab.sim.sensors import CameraCfg
from embodichain.lab.sim.solvers import PinkSolverCfg
from embodichain.data import get_data_path
from embodichain.utils import logger


def main():
    """Main function to create and run the simulation scene."""

    parser = argparse.ArgumentParser(
        description="Create a simulation scene with SimulationManager"
    )
    parser.add_argument(
        "--device", type=str, default="cpu", help="Simulation device (cuda or cpu)"
    )
    parser.add_argument(
        "--enable_rt",
        action="store_true",
        default=False,
        help="Enable ray tracing for better visuals",
    )
    args = parser.parse_args()

    # Configure the simulation
    sim_cfg = SimulationManagerCfg(
        width=1920,
        height=1080,
        physics_dt=1.0 / 100.0,
        sim_device=args.device,
        enable_rt=args.enable_rt,
    )

    sim = SimulationManager(sim_cfg)
    sim.set_manual_update(False)

    # Get DexForce W1 URDF path
    urdf_path = get_data_path(
        "DexforceW1V021_INDUSTRIAL_DH_PGC_GRIPPER_M/DexforceW1V021.urdf"
    )

    # Create DexForce W1 robot
    robot_cfg = RobotCfg(
        uid="w1_gizmo_test",
        urdf_cfg=URDFCfg(
            components=[{"component_type": "humanoid", "urdf_path": urdf_path}]
        ),
        control_parts={"left_arm": ["LEFT_J[1-7]"], "right_arm": ["RIGHT_J[1-7]"]},
        solver_cfg={
            "left_arm": PinkSolverCfg(
                urdf_path=urdf_path,
                end_link_name="left_ee",
                root_link_name="left_arm_base",
                pos_eps=1e-2,
                rot_eps=5e-2,
                max_iterations=300,
                dt=0.1,
            ),
            "right_arm": PinkSolverCfg(
                urdf_path=urdf_path,
                end_link_name="right_ee",
                root_link_name="right_arm_base",
                pos_eps=1e-2,
                rot_eps=5e-2,
                max_iterations=300,
                dt=0.1,
            ),
        },
        drive_pros=JointDrivePropertiesCfg(
            stiffness={"LEFT_J[1-7]": 1e4, "RIGHT_J[1-7]": 1e4},
            damping={"LEFT_J[1-7]": 1e3, "RIGHT_J[1-7]": 1e3},
        ),
    )
    robot = sim.add_robot(cfg=robot_cfg)

    # Set initial joint positions for both arms
    left_arm_qpos = torch.tensor(
        [
            [0, 0, -np.pi / 4, np.pi / 4, -np.pi / 2, 0.0, np.pi / 4, 0.0]
        ],  # WAIST + LEFT_J[1-7]
        dtype=torch.float32,
        device="cpu",
    )
    right_arm_qpos = torch.tensor(
        [
            [0, 0, np.pi / 4, -np.pi / 4, np.pi / 2, 0.0, -np.pi / 4, 0.0]
        ],  # WAIST + RIGHT_J[1-7]
        dtype=torch.float32,
        device="cpu",
    )

    left_joint_ids = robot.get_joint_ids("left_arm")
    right_joint_ids = robot.get_joint_ids("right_arm")

    robot.set_qpos(qpos=left_arm_qpos, joint_ids=left_joint_ids)
    robot.set_qpos(qpos=right_arm_qpos, joint_ids=right_joint_ids)

    # Create a rigid object (cube) positioned to the side of the robot
    cube_cfg = RigidObjectCfg(
        uid="interactive_cube",
        shape=CubeCfg(size=[0.1, 0.1, 0.1]),
        body_type="kinematic",
        attrs=RigidBodyAttributesCfg(
            mass=1.0,
            dynamic_friction=0.5,
            static_friction=0.5,
            restitution=0.1,
        ),
        init_pos=[1.0, 0.0, 0.5],  # Position to the side of the robot
    )
    cube = sim.add_rigid_object(cube_cfg)

    camera_cfg = CameraCfg(
        uid="scene_camera",
        width=640,
        height=480,
        intrinsics=(320, 320, 320, 240),  # fx, fy, cx, cy
        near=0.1,
        far=10.0,
        enable_color=True,
        enable_depth=True,
        extrinsics=CameraCfg.ExtrinsicsCfg(
            eye=(2.0, 2.0, 2.0),
            target=(0.0, 0.0, 0.0),
            up=(0.0, 0.0, 1.0),
        ),
    )
    camera = sim.add_sensor(sensor_cfg=camera_cfg)

    # Enable gizmo for all assets after all are created and initialized
    sim.enable_gizmo(uid="w1_gizmo_test", control_part="left_arm")
    if not sim.has_gizmo("w1_gizmo_test", control_part="left_arm"):
        logger.log_error("Failed to enable left arm gizmo!")
        return

    sim.enable_gizmo(uid="w1_gizmo_test", control_part="right_arm")
    if not sim.has_gizmo("w1_gizmo_test", control_part="right_arm"):
        logger.log_error("Failed to enable right arm gizmo!")
        return

    sim.enable_gizmo(uid="interactive_cube")
    if not sim.has_gizmo("interactive_cube"):
        logger.log_error("Failed to enable gizmo for cube!")
        return

    sim.enable_gizmo(uid="scene_camera")
    if not sim.has_gizmo("scene_camera"):
        logger.log_error("Failed to enable gizmo for camera!")
        return

    sim.open_window()

    logger.log_info("Gizmo Scene example started!")
    logger.log_info("Four gizmos are active in the scene:")
    logger.log_info("1. Left arm gizmo - Use to drag the left arm end-effector (EE)")
    logger.log_info("2. Right arm gizmo - Use to drag the right arm end-effector (EE)")
    logger.log_info("3. Cube gizmo - Use to drag and position the cube")
    logger.log_info("4. Camera gizmo - Use to drag and orient the camera")
    logger.log_info("Press Ctrl+C to stop the simulation")

    run_simulation(sim)


def run_simulation(sim: SimulationManager):
    step_count = 0
    # Get the camera instance by uid
    camera = sim.get_sensor("scene_camera")
    try:
        last_time = time.time()
        last_step = 0
        while True:
            time.sleep(0.033)  # 30Hz
            sim.update_gizmos()
            step_count += 1

            # Display camera view in a window every 5 steps
            if camera is not None and step_count % 5 == 0:
                camera.update()
                data = camera.get_data()
                if "color" in data:
                    rgb_image = data["color"].cpu().numpy()[0, :, :, :3]
                    bgr_image = cv2.cvtColor(rgb_image, cv2.COLOR_RGB2BGR)
                    cv2.putText(
                        bgr_image,
                        "Press 'h' to toggle camera gizmo visibility",
                        (10, 30),
                        cv2.FONT_HERSHEY_SIMPLEX,
                        0.6,
                        (0, 255, 0),
                        2,
                    )
                    cv2.imshow("Camera Sensor View", bgr_image)
                    key = cv2.waitKey(1) & 0xFF
                    if key == ord("h"):
                        # Toggle the camera gizmo visibility using SimulationManager API
                        sim.toggle_gizmo_visibility("scene_camera")

            if step_count % 100 == 0:
                current_time = time.time()
                elapsed = current_time - last_time
                fps = (
                    sim.num_envs * (step_count - last_step) / elapsed
                    if elapsed > 0
                    else 0
                )
                logger.log_info(f"Simulation step: {step_count}, FPS: {fps:.2f}")
                last_time = current_time
                last_step = step_count
    except KeyboardInterrupt:
        logger.log_info("\nStopping simulation...")
    finally:
        cv2.destroyAllWindows()
        sim.destroy()
        logger.log_info("Simulation terminated successfully")


if __name__ == "__main__":
    main()



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import torch
import numpy as np
from dataclasses import dataclass
from scipy.spatial.transform import Rotation, Slerp
from enum import Enum
from typing import Union, List

from embodichain.utils import logger


__all__ = [
    "TrajectorySampleMethod",
    "MovePart",
    "MoveType",
    "PlanState",
    "PlanResult",
    "calculate_point_allocations",
    "interpolate_xpos",
]


class TrajectorySampleMethod(Enum):
    r"""Enumeration for different trajectory sampling methods.

    This enum defines various methods for sampling trajectories,
    providing meaningful names for different sampling strategies.
    """

    TIME = "time"
    """Sample based on time intervals."""

    QUANTITY = "quantity"
    """Sample based on a specified number of points."""

    DISTANCE = "distance"
    """Sample based on distance intervals."""

    @classmethod
    def from_str(
        cls, value: Union[str, "TrajectorySampleMethod"]
    ) -> "TrajectorySampleMethod":
        if isinstance(value, cls):
            return value
        try:
            return cls[value.upper()]
        except KeyError:
            valid_values = [e.name for e in cls]
            logger.log_error(
                f"Invalid version '{value}'. Valid values are: {valid_values}",
                ValueError,
            )

    def __str__(self):
        """Override string representation for better readability."""
        return self.value.capitalize()


class MovePart(Enum):
    r"""Enumeration for different robot parts to move.

    Defines robot part selection for motion planning.

    Attributes:
        LEFT (int): left arm or end-effector.
        RIGHT (int): right arm or end-effector.
        BOTH (int): both arms or end-effectors.
        TORSO (int): torso for humanoid robot.
        ALL (int): all joints of the robot (joint control only).
    """

    LEFT = 0  # left arm|eef
    RIGHT = 1  # right arm|eef
    BOTH = 2  # left arm|eef and right arm|eef
    TORSO = 3  # torso for humanoid robot
    ALL = 4  # all joints of the robot. Only for joint control.


class MoveType(Enum):
    r"""Enumeration for different types of movements.

    Defines movement types for robot planning.

    Attributes:
        TOOL (int): Tool open or close.
        EEF_MOVE (int): Move end-effector to target pose (IK + trajectory).
        JOINT_MOVE (int): Move joints to target angles (trajectory planning).
        SYNC (int): Synchronized left/right arm movement (dual-arm robots).
        PAUSE (int): Pause for specified duration (see PlanState.pause_seconds).
    """

    TOOL = 0  # Tool open or close
    EEF_MOVE = 1  # Move the end-effector to a target pose (xpos) using IK and trajectory planning
    JOINT_MOVE = (
        2  # Directly move joints to target angles (qpos) using trajectory planning
    )
    SYNC = 3  # Synchronized left and right arm movement (for dual-arm robots)
    PAUSE = 4  # Pause for a specified duration (use pause_seconds in PlanState)


@dataclass
class PlanResult:
    r"""Data class representing the result of a motion plan."""

    success: bool | torch.Tensor = False
    """Whether planning succeeded."""

    xpos_list: torch.Tensor | None = None
    """End-effector poses along trajectory with shape `(N, 4, 4)`."""

    positions: torch.Tensor | None = None
    """Joint positions along trajectory with shape `(N, DOF)`."""

    velocities: torch.Tensor | None = None
    """Joint velocities along trajectory with shape `(N, DOF)`."""

    accelerations: torch.Tensor | None = None
    """Joint accelerations along trajectory with shape `(N, DOF)`."""

    dt: torch.Tensor | None = None
    """Time duration between each point with shape `(N,)`."""

    duration: float | torch.Tensor = 0.0
    """Total trajectory duration in seconds."""


@dataclass
class PlanState:
    r"""Data class representing the state for a motion plan."""

    move_type: MoveType = MoveType.JOINT_MOVE
    """Type of movement used by the plan."""

    move_part: MovePart = MovePart.LEFT
    """Robot part that should move."""

    xpos: torch.Tensor | None = None
    """Target TCP pose (4x4 matrix) for `MoveType.EEF_MOVE`."""

    qpos: torch.Tensor | None = None
    """Target joint angles for `MoveType.JOINT_MOVE` with shape `(DOF,)`."""

    qvel: torch.Tensor | None = None
    """Target joint velocities for `MoveType.JOINT_MOVE` with shape `(DOF,)`."""

    qacc: torch.Tensor | None = None
    """Target joint accelerations for `MoveType.JOINT_MOVE` with shape `(DOF,)`."""

    is_open: bool = True
    """For `MoveType.TOOL`, indicates whether to open (`True`) or close (`False`) the tool."""

    is_world_coordinate: bool = True
    """`True` if the target pose is in world coordinates, `False` if relative to the current pose."""

    pause_seconds: float = 0.0
    """Duration of a pause when `move_type` is `MoveType.PAUSE`."""


def interpolate_xpos(
    current_xpos: np.ndarray, target_xpos: np.ndarray, num_samples: int
) -> np.ndarray:
    """Interpolate between two poses using vectorized Slerp + linear translation."""
    num_samples = max(2, int(num_samples))

    interp_ratios = np.linspace(0.0, 1.0, num_samples)
    slerp = Slerp(
        [0.0, 1.0],
        Rotation.from_matrix([current_xpos[:3, :3], target_xpos[:3, :3]]),
    )
    interp_rots = slerp(interp_ratios).as_matrix()
    interp_trans = (1.0 - interp_ratios[:, None]) * current_xpos[:3, 3] + interp_ratios[
        :, None
    ] * target_xpos[:3, 3]

    interp_poses = np.repeat(np.eye(4)[None, :, :], num_samples, axis=0)
    interp_poses[:, :3, :3] = interp_rots
    interp_poses[:, :3, 3] = interp_trans
    return interp_poses


def calculate_point_allocations(
    xpos_list: torch.Tensor | np.ndarray,
    step_size: float = 0.002,
    angle_step: float = np.pi / 90,
    device: torch.device = torch.device("cpu"),
) -> List[int]:
    """Calculate interpolation points for each segment with vectorized tensor ops."""
    if not isinstance(xpos_list, torch.Tensor):
        xpos_tensor = torch.as_tensor(
            np.asarray(xpos_list), dtype=torch.float32, device=device
        )
    else:
        xpos_tensor = xpos_list.to(dtype=torch.float32, device=device)

    if xpos_tensor.dim() != 3 or xpos_tensor.shape[0] < 2:
        return []

    start_poses = xpos_tensor[:-1]  # [N-1, 4, 4]
    end_poses = xpos_tensor[1:]  # [N-1, 4, 4]

    pos_dists = torch.norm(end_poses[:, :3, 3] - start_poses[:, :3, 3], dim=-1)
    pos_points = torch.clamp((pos_dists / step_size).int(), min=1)

    rel_rot = torch.matmul(
        start_poses[:, :3, :3].transpose(-1, -2), end_poses[:, :3, :3]
    )
    trace = rel_rot[:, 0, 0] + rel_rot[:, 1, 1] + rel_rot[:, 2, 2]
    cos_angle = torch.clamp((trace - 1.0) / 2.0, -1.0 + 1e-6, 1.0 - 1e-6)
    angles = torch.acos(cos_angle)
    rot_points = torch.clamp((angles / angle_step).int(), min=1)

    return torch.maximum(pos_points, rot_points).tolist()



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import os
import torch
import pytest

from embodichain.lab.sim import SimulationManager, SimulationManagerCfg
from embodichain.lab.sim.objects import RigidObjectGroup
from embodichain.lab.sim.cfg import RigidObjectGroupCfg, RigidObjectCfg
from embodichain.lab.sim.shapes import MeshCfg
from embodichain.data import get_data_path
from dexsim.types import ActorType

DUCK_PATH = "ToyDuck/toy_duck.glb"
TABLE_PATH = "ShopTableSimple/shop_table_simple.ply"
NUM_ARENAS = 4
Z_TRANSLATION = 2.0


class BaseRigidObjectGroupTest:
    """Shared test logic for CPU and CUDA."""

    def setup_simulation(self, sim_device):
        config = SimulationManagerCfg(
            headless=True, sim_device=sim_device, num_envs=NUM_ARENAS
        )
        self.sim = SimulationManager(config)

        duck_path = get_data_path(DUCK_PATH)
        assert os.path.isfile(duck_path)
        table_path = get_data_path(TABLE_PATH)
        assert os.path.isfile(table_path)

        cfg_dict = {
            "uid": "group",
            "rigid_objects": {
                "duck1": {
                    "shape": {
                        "shape_type": "Mesh",
                        "fpath": duck_path,
                    },
                },
                "duck2": {
                    "shape": {
                        "shape_type": "Mesh",
                        "fpath": duck_path,
                    },
                },
            },
        }
        self.obj_group: RigidObjectGroup = self.sim.add_rigid_object_group(
            cfg=RigidObjectGroupCfg.from_dict(cfg_dict)
        )

        if sim_device == "cuda" and self.sim.is_use_gpu_physics:
            self.sim.init_gpu_physics()

        self.sim.enable_physics(True)

    def test_local_pose_behavior(self):

        # Set initial poses
        pose_duck1 = torch.eye(4, device=self.sim.device)
        pose_duck1[2, 3] = Z_TRANSLATION
        pose_duck1 = pose_duck1.unsqueeze(0).repeat(NUM_ARENAS, 1, 1)

        pose_duck2 = torch.eye(4, device=self.sim.device)
        pose_duck2[2, 3] = Z_TRANSLATION
        pose_duck2 = pose_duck2.unsqueeze(0).repeat(NUM_ARENAS, 1, 1)

        combined_pose = torch.stack([pose_duck1, pose_duck2], dim=1)

        self.obj_group.set_local_pose(combined_pose)
        group_pos = self.obj_group.get_local_pose()[..., :3]
        assert torch.allclose(
            group_pos,
            combined_pose[..., :3, 3],
            atol=1e-5,
        ), "FAIL: Local poses do not match after setting."

    def test_get_user_ids(self):
        """Test get_user_ids method."""
        user_ids = self.obj_group.get_user_ids()

        assert user_ids.shape == (NUM_ARENAS, self.obj_group.num_objects), (
            f"Unexpected user_ids shape: {user_ids.shape}, "
            f"expected ({NUM_ARENAS}, {self.obj_group.num_objects})"
        )

    def test_remove(self):
        self.sim.remove_asset(self.obj_group.uid)

        assert (
            self.obj_group.uid not in self.sim.asset_uids
        ), "Object group UID still present after removal"

    def test_set_physical_visible(self):
        self.obj_group.set_physical_visible(visible=True)
        self.obj_group.set_physical_visible(visible=False)

    def test_set_visible(self):
        self.obj_group.set_visible(visible=True)
        self.obj_group.set_visible(visible=False)

    def teardown_method(self):
        """Clean up resources after each test method."""
        self.sim.destroy()


class TestRigidObjectGroupCPU(BaseRigidObjectGroupTest):
    def setup_method(self):
        self.setup_simulation("cpu")


# TODO: Fix CUDA tests issue.
@pytest.mark.skip(reason="Skipping CUDA tests temporarily")
class TestRigidObjectGroupCUDA(BaseRigidObjectGroupTest):
    def setup_method(self):
        self.setup_simulation("cuda")


if __name__ == "__main__":
    # pytest.main(["-s", __file__])
    test = TestRigidObjectGroupCPU()
    test.setup_method()
    test.test_local_pose_behavior()



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

"""Dataset functors for collecting and saving episode data."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Optional, Union

import numpy as np
import gymnasium as gym
import torch
import tqdm

from tensordict import TensorDict

from embodichain.utils import logger
from embodichain.data.constants import EMBODICHAIN_DEFAULT_DATASET_ROOT
from embodichain.lab.gym.utils.misc import is_stereocam
from embodichain.utils.utility import get_right_name
from embodichain.data.enum import JointType
from .manager_base import Functor
from .cfg import DatasetFunctorCfg

if TYPE_CHECKING:
    from embodichain.lab.gym.envs import EmbodiedEnv

try:
    from lerobot.datasets.lerobot_dataset import LeRobotDataset

    LEROBOT_AVAILABLE = True
    __all__ = ["LeRobotRecorder"]
except ImportError:
    LEROBOT_AVAILABLE = False
    __all__ = []


class LeRobotRecorder(Functor):
    """Functor for recording episodes in LeRobot format.

    This functor handles:
    - Recording observation-action pairs during episodes
    - Converting data to LeRobot format
    - Saving episodes when they complete
    """

    def __init__(self, cfg: DatasetFunctorCfg, env: EmbodiedEnv):
        """Initialize the LeRobot dataset recorder.

        Args:
            cfg: Functor configuration containing params:
                - save_path: Root directory for saving datasets
                - robot_meta: Robot metadata for dataset
                - instruction: Optional task instruction
                - extra: Optional extra metadata
                - use_videos: Whether to save videos
                - image_writer_threads: Number of threads for image writing
                - image_writer_processes: Number of processes for image writing
            env: The environment instance
        """
        if not LEROBOT_AVAILABLE:
            logger.log_error(
                "LeRobot is not installed. Please install it with: pip install lerobot"
            )

        super().__init__(cfg, env)

        # Extract parameters from cfg.params
        params = cfg.params

        # Required parameters
        self.lerobot_data_root = params.get(
            "save_path", EMBODICHAIN_DEFAULT_DATASET_ROOT
        )
        self.robot_meta = params.get("robot_meta", {})

        # Optional parameters
        self.instruction = params.get("instruction", None)
        self.extra = params.get("extra", {})

        # Experimental parameters for extra episode info saving.
        self.extra_episode_info = self.extra.get("episode_info", {})
        self.extra_episode_info_buffer = {}
        self.use_videos = params.get("use_videos", False)

        # LeRobot dataset instance
        self.dataset: Optional[LeRobotDataset] = None
        self.dataset_full_path: Optional[Path] = None

        # Tracking
        self.total_time: float = 0.0
        self.curr_episode: int = 0

        # Initialize dataset
        self._initialize_dataset()

    @property
    def dataset_path(self) -> str:
        """Path to the dataset directory."""
        return (
            str(self.dataset_full_path) if self.dataset_full_path else "Not initialized"
        )

    def __call__(
        self,
        env: EmbodiedEnv,
        env_ids: Union[torch.Tensor, None],
        save_path: Optional[str] = None,
        robot_meta: Optional[Dict] = None,
        instruction: Optional[str] = None,
        extra: Optional[Dict] = None,
        use_videos: bool = False,
    ) -> None:
        """Main entry point for the recorder functor.

        This method is called by DatasetManager.apply(mode="save") to save completed episodes.
        It reads data from the environment's episode buffers.

        Args:
            env: The environment instance.
            env_ids: Environment IDs to save. If None, attempts to save all environments.
        """
        # If env_ids is None, check all environments for completed episodes
        if env_ids is None:
            env_ids = torch.arange(env.num_envs, device=env.device)
        elif isinstance(env_ids, (list, range)):
            env_ids = torch.tensor(list(env_ids), device=env.device)

        # Save episodes for specified environments
        if len(env_ids) > 0:
            self._save_episodes(env_ids)

    def _save_episodes(
        self,
        env_ids: torch.Tensor,
    ) -> None:
        """Save completed episodes for specified environments."""
        task = self.instruction.get("lang", "unknown_task")

        # Process each environment
        for env_id in env_ids.cpu().tolist():
            # Get buffer for this environment (already contains single-env data)
            obs_list = self._env.rollout_buffer["obs"][
                env_id, : self._env.current_rollout_step
            ]
            action_list = self._env.rollout_buffer["actions"][
                env_id, : self._env.current_rollout_step
            ]

            if len(obs_list) == 0:
                logger.log_warning(f"No episode data to save for env {env_id}")
                continue

            # Align obs and action
            if len(obs_list) > len(action_list):
                obs_list = obs_list[:-1]

            # Update metadata
            extra_info = self.extra.copy() if self.extra else {}
            fps = self.dataset.meta.info.get("fps", 30)
            current_episode_time = len(obs_list) / fps if fps > 0 else 0

            episode_extra_info = extra_info.copy()
            self.total_time += current_episode_time
            episode_extra_info["total_time"] = self.total_time
            self._save_extra_episode_meta_info(env_id)

            try:
                for obs, action in tqdm.tqdm(
                    zip(obs_list, action_list),
                    total=len(obs_list),
                    desc=f"Converting env {env_id} episode to LeRobot format",
                ):
                    frame = self._convert_frame_to_lerobot(obs, action, task)
                    self.dataset.add_frame(frame)

                self.dataset.save_episode()

                logger.log_info(
                    f"[LeRobotRecorder] Saved dataset to: {self.dataset_path}\n"
                    f"  Episode {self.curr_episode} (env {env_id}): {len(obs_list)} frames"
                )

                self.curr_episode += 1
            except Exception as e:
                logger.log_error(f"Failed to save episode {env_id}: {e}")

    def _save_extra_episode_meta_info(self, env_id: int) -> None:
        """Save extra episode meta info for a specific environment ID."""

        curr_extra_episode_info = {}
        if self.extra_episode_info:
            for key, attr_list in self.extra_episode_info.items():
                if key == "rigid_object_physics_attributes":
                    rigid_obj_list = self._env.sim.get_rigid_object_uid_list()
                    for obj_uid in rigid_obj_list:
                        curr_extra_episode_info[obj_uid] = {}
                        obj = self._env.sim.get_rigid_object(obj_uid)
                        for attr in attr_list:
                            if attr == "mass":
                                curr_extra_episode_info[obj_uid]["mass"] = round(
                                    obj.get_mass(env_ids=[env_id]).squeeze_().item(), 5
                                )

        self.extra_episode_info_buffer[self.curr_episode] = curr_extra_episode_info
        self._update_dataset_info(
            {"extra_episode_info": self.extra_episode_info_buffer}
        )

    def finalize(self) -> Optional[str]:
        """Finalize the dataset."""
        # Save any remaining episodes
        if self._env.current_rollout_step > 0:
            active_env_ids = torch.arange(self._env.num_envs, device=self._env.device)
            self._save_episodes(active_env_ids)

        try:
            if self.dataset is not None:
                self.dataset.finalize()
                logger.log_info(
                    f"[LeRobotRecorder] Dataset finalized successfully\n"
                    f"  Path: {self.dataset_path}\n"
                    f"  Total episodes: {self.curr_episode}\n"
                    f"  Total time: {self.total_time:.2f}s"
                )
                return self.dataset_path
        except Exception as e:
            logger.log_error(f"[LeRobotRecorder] Failed to finalize dataset: {e}")

        return None

    def _initialize_dataset(self) -> None:
        """Initialize the LeRobot dataset."""
        robot_type = self.robot_meta.get("robot_type", "robot")
        scene_type = self.extra.get("scene_type", "scene")
        task_description = self.extra.get("task_description", "task")

        robot_type = str(robot_type).lower().replace(" ", "_")
        task_description = str(task_description).lower().replace(" ", "_")

        # Use lerobot_data_root from __init__
        lerobot_data_root = Path(self.lerobot_data_root)

        # Generate dataset folder name with auto-incrementing suffix
        base_name = f"{robot_type}_{scene_type}_{task_description}"

        # Find the next available sequence number by checking existing folders
        existing_dirs = list(lerobot_data_root.glob(f"{base_name}_*"))
        if not existing_dirs:
            dataset_id = 0
        else:
            # Extract sequence numbers from existing directories
            max_id = -1
            for dir_path in existing_dirs:
                suffix = dir_path.name[len(base_name) + 1 :]  # +1 for underscore
                if suffix.isdigit():
                    max_id = max(max_id, int(suffix))
            dataset_id = max_id + 1

        # Format dataset name with zero-padding (3 digits: 000, 001, 002, ...)
        dataset_name = f"{base_name}_{dataset_id:03d}"

        # LeRobot's root parameter is the COMPLETE dataset path (not parent directory)
        self.dataset_full_path = lerobot_data_root / dataset_name

        fps = self.robot_meta.get("control_freq", 30)
        features = self._build_features()

        self.dataset = LeRobotDataset.create(
            repo_id=dataset_name,
            fps=fps,
            root=str(self.dataset_full_path),
            robot_type=robot_type,
            features=features,
            use_videos=self.use_videos,
            metadata_buffer_size=1,
        )
        logger.log_info(f"Created LeRobot dataset at: {self.dataset_full_path}")

    def _build_features(self) -> Dict:
        """Build LeRobot features dict."""
        features = {}

        state_dim = len(self._env.active_joint_ids)
        # Create joint names.
        joint_names = [
            self._env.robot.joint_names[i] for i in self._env.active_joint_ids
        ]

        features["observation.qpos"] = {
            "dtype": "float32",
            "shape": (state_dim,),
            "names": joint_names,
        }
        features["observation.qvel"] = {
            "dtype": "float32",
            "shape": (state_dim,),
            "names": joint_names,
        }
        features["observation.qf"] = {
            "dtype": "float32",
            "shape": (state_dim,),
            "names": joint_names,
        }

        # Use full qpos dimension for action (includes gripper)
        action_dim = state_dim
        features["action"] = {
            "dtype": "float32",
            "shape": (action_dim,),
            "names": joint_names,
        }

        # Setup sensor observation features based env.observation.sensor
        if self._env.has_sensors:
            sensor_obs_space: dict = self._env.single_observation_space["sensor"]

            for sensor_name, value in sensor_obs_space.items():
                sensor = self._env.get_sensor(sensor_name)
                is_stereo = is_stereocam(sensor)

                for frame_name, space in value.items():
                    # TODO: Support depth (uint16) and mask (also uint16 or uint8)
                    if frame_name not in ["color", "color_right"]:
                        logger.log_error(
                            f"Only support 'color' frame for vision sensors, but got '{frame_name}' in sensor '{sensor_name}'"
                        )

                    features[f"{sensor_name}.{frame_name}"] = {
                        "dtype": "video" if self.use_videos else "image",
                        "shape": (sensor.cfg.height, sensor.cfg.width, 3),
                        "names": ["height", "width", "channel"],
                    }

                    if is_stereo:
                        features[f"{sensor_name}.{frame_name}_right"] = {
                            "dtype": "video" if self.use_videos else "image",
                            "shape": (sensor.cfg.height, sensor.cfg.width, 3),
                            "names": ["height", "width", "channel"],
                        }

        # TODO: The extra observation features are supposed to be defined in a flattened way in the observation space.
        # Lerobot requires a flat feature dict, so we may need to support nested dicts to flatten dict conversion in the future.
        # Add any extra features specified in observation space excluding 'robot' and 'sensor'
        for key, space in self._env.single_observation_space.items():
            if key in ["robot", "sensor"]:
                continue

            if isinstance(space, gym.spaces.Dict):
                logger.log_warning(
                    f"Nested Dict observation space for key '{key}' is not directly supported. "
                    f"Please flatten it or specify features manually. Skipping '{key}'."
                )
                continue

            names = key
            if "vel" in key:
                names = ["lin_x", "lin_y", "lin_z", "ang_x", "ang_y", "ang_z"]
            elif "pose" in key:
                names = ["x", "y", "z", "qw", "qx", "qy", "qz"]

            features[f"observation.{key}"] = {
                "dtype": str(space.dtype),
                "shape": space.shape,
                "names": names,
            }

        return features

    def _convert_frame_to_lerobot(
        self, obs: TensorDict, action: TensorDict | torch.Tensor, task: str
    ) -> Dict:
        """Convert a single frame to LeRobot format.

        Args:
            obs: Single environment observation (already extracted from batch)
            action: Single environment action (already extracted from batch)
            task: Task name

        Returns:
            Frame dict in LeRobot format with numpy arrays
        """
        frame = {"task": task}

        if self._env.has_sensors:
            sensor_obs_space: dict = self._env.single_observation_space["sensor"]

            # Add images
            for sensor_name, value in sensor_obs_space.items():
                sensor = self._env.get_sensor(sensor_name)
                is_stereo = is_stereocam(sensor)

                color_data = obs["sensor"][sensor_name]["color"]
                color_img = color_data[:, :, :3].cpu()
                frame[f"{sensor_name}.color"] = color_img

                if is_stereo:
                    color_right_data = obs["sensor"][sensor_name]["color_right"]
                    color_right_img = color_right_data[:, :, :3].cpu()
                    frame[f"{sensor_name}.color_right"] = color_right_img

        # Add state
        frame["observation.qpos"] = obs["robot"]["qpos"].cpu()
        frame["observation.qvel"] = obs["robot"]["qvel"].cpu()
        frame["observation.qf"] = obs["robot"]["qf"].cpu()

        # Add extra observation features if they exist
        for key in obs.keys():
            if key in ["robot", "sensor"]:
                continue

            frame[f"observation.{key}"] = obs[key].cpu()

        # Add action.
        if isinstance(action, torch.Tensor):
            action_data = action.cpu()
        elif isinstance(action, TensorDict):
            # Extract qpos from action dict
            action_tensor = action.get("qpos", action.get("delta_qpos", None))
            if action_tensor is None:
                # Fallback to first tensor value
                for v in action.values():
                    if isinstance(v, (torch.Tensor, np.ndarray)):
                        action_tensor = v
                        break

            if isinstance(action_tensor, torch.Tensor):
                action_data = action_tensor.cpu()

        frame["action"] = action_data

        return frame

    def _update_dataset_info(self, updates: dict) -> bool:
        """Update dataset metadata."""
        if self.dataset is None:
            logger.log_error("LeRobotDataset not initialized.")
            return False

        try:
            self.dataset.meta.info.update(updates)
            return True
        except Exception as e:
            logger.log_error(f"Failed to update dataset info: {e}")
            return False



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

from __future__ import annotations

from typing import Callable

import torch
from tensordict import TensorDict

from embodichain.agents.rl.utils import dict_to_tensordict, flatten_dict_observation
from .base import BaseCollector

__all__ = ["SyncCollector"]


class SyncCollector(BaseCollector):
    """Synchronously collect rollouts from a vectorized environment."""

    def __init__(
        self,
        env,
        policy,
        device: torch.device,
        reset_every_rollout: bool = False,
    ) -> None:
        self.env = env
        self.policy = policy
        self.device = device
        self.reset_every_rollout = reset_every_rollout
        self._supports_shared_rollout = hasattr(self.env, "set_rollout_buffer")
        self.obs_td = self._reset_env()

    @torch.no_grad()
    def collect(
        self,
        num_steps: int,
        rollout: TensorDict | None = None,
        on_step_callback: Callable[[TensorDict, dict], None] | None = None,
    ) -> TensorDict:
        self.policy.train()
        if self.reset_every_rollout:
            self.obs_td = self._reset_env()

        if rollout is None:
            raise ValueError(
                "SyncCollector.collect() requires a preallocated rollout TensorDict."
            )
        if tuple(rollout.batch_size) != (self.env.num_envs, num_steps + 1):
            raise ValueError(
                "Preallocated rollout batch size mismatch: "
                f"expected ({self.env.num_envs}, {num_steps + 1}), got {tuple(rollout.batch_size)}."
            )
        self._validate_rollout(rollout, num_steps)
        if self._supports_shared_rollout:
            self.env.set_rollout_buffer(rollout)

        initial_obs = flatten_dict_observation(self.obs_td)
        rollout["obs"][:, 0] = initial_obs
        for step_idx in range(num_steps):
            step_td = TensorDict(
                {"obs": rollout["obs"][:, step_idx]},
                batch_size=[rollout.batch_size[0]],
                device=self.device,
            )
            step_td = self.policy.get_action(step_td)

            next_obs, reward, terminated, truncated, env_info = self.env.step(
                self._to_action_dict(step_td["action"])
            )
            next_obs_td = dict_to_tensordict(next_obs, self.device)
            self._write_step(
                rollout=rollout,
                step_idx=step_idx,
                step_td=step_td,
            )
            if not self._supports_shared_rollout:
                self._write_env_step(
                    rollout=rollout,
                    step_idx=step_idx,
                    reward=reward,
                    terminated=terminated,
                    truncated=truncated,
                )
            rollout["obs"][:, step_idx + 1] = flatten_dict_observation(next_obs_td)

            if on_step_callback is not None:
                on_step_callback(rollout[:, step_idx], env_info)

            self.obs_td = next_obs_td

        self._attach_final_value(rollout)
        return rollout

    def _attach_final_value(self, rollout: TensorDict) -> None:
        """Populate the bootstrap value for the final observed state."""
        final_obs = rollout["obs"][:, -1]
        last_next_td = TensorDict(
            {"obs": final_obs},
            batch_size=[rollout.batch_size[0]],
            device=self.device,
        )
        self.policy.get_value(last_next_td)
        rollout["value"][:, -1] = last_next_td["value"]

    def _reset_env(self) -> TensorDict:
        obs, _ = self.env.reset()
        return dict_to_tensordict(obs, self.device)

    def _to_action_dict(self, action: torch.Tensor) -> dict[str, torch.Tensor]:
        am = getattr(self.env, "action_manager", None)
        action_type = (
            am.action_type if am else getattr(self.env, "action_type", "delta_qpos")
        )
        return {action_type: action}

    def _write_step(
        self,
        rollout: TensorDict,
        step_idx: int,
        step_td: TensorDict,
    ) -> None:
        """Write policy-side fields for one transition into the shared rollout TensorDict."""
        rollout["action"][:, step_idx] = step_td["action"]
        rollout["sample_log_prob"][:, step_idx] = step_td["sample_log_prob"]
        rollout["value"][:, step_idx] = step_td["value"]

    def _write_env_step(
        self,
        rollout: TensorDict,
        step_idx: int,
        reward: torch.Tensor,
        terminated: torch.Tensor,
        truncated: torch.Tensor,
    ) -> None:
        """Populate transition-side fields when the environment does not own the rollout."""
        done = terminated | truncated
        rollout["reward"][:, step_idx] = reward.to(self.device)
        rollout["done"][:, step_idx] = done.to(self.device)
        rollout["terminated"][:, step_idx] = terminated.to(self.device)
        rollout["truncated"][:, step_idx] = truncated.to(self.device)

    def _validate_rollout(self, rollout: TensorDict, num_steps: int) -> None:
        """Validate rollout layout expected by the collector."""
        expected_shapes = {
            "obs": (self.env.num_envs, num_steps + 1, self.policy.obs_dim),
            "action": (self.env.num_envs, num_steps + 1, self.policy.action_dim),
            "sample_log_prob": (self.env.num_envs, num_steps + 1),
            "value": (self.env.num_envs, num_steps + 1),
            "reward": (self.env.num_envs, num_steps + 1),
            "done": (self.env.num_envs, num_steps + 1),
            "terminated": (self.env.num_envs, num_steps + 1),
            "truncated": (self.env.num_envs, num_steps + 1),
        }
        for key, expected_shape in expected_shapes.items():
            actual_shape = tuple(rollout[key].shape)
            if actual_shape != expected_shape:
                raise ValueError(
                    f"Preallocated rollout field '{key}' shape mismatch: "
                    f"expected {expected_shape}, got {actual_shape}."
                )



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import pytest
import torch
from embodichain.lab.sim import SimulationManager, SimulationManagerCfg
from embodichain.lab.sim.cfg import LightCfg


class TestLight:
    def setup_method(self):
        # Setup SimulationManager
        config = SimulationManagerCfg(headless=True, sim_device="cpu", num_envs=10)
        self.sim = SimulationManager(config)

        # Create batch of lights
        cfg_dict = {
            "light_type": "point",
            "color": [0.1, 0.1, 0.1],
            "radius": 10.0,
            "position": [0.0, 0.0, 2.0],
            "uid": "point_light",
        }
        self.light = self.sim.add_light(cfg=LightCfg.from_dict(cfg_dict))

    def test_set_color_with_env_ids(self):
        """Test set_color with and without env_ids."""
        base_color = torch.tensor([0.1, 0.1, 0.1], device=self.sim.device)

        # Set for all environments
        try:
            self.light.set_color(base_color)
        except Exception as e:
            pytest.fail(f"Failed to set color for all envs: {e}")

        # Set for specific envs
        env_ids = [1, 3, 5]
        new_color = torch.tensor([0.9, 0.8, 0.7], device=self.sim.device)
        try:
            self.light.set_color(new_color, env_ids=env_ids)
        except Exception as e:
            pytest.fail(f"Failed to set color for env_ids={env_ids}: {e}")

    def test_set_falloff_with_env_ids(self):
        """Test set_falloff with and without env_ids."""
        base_val = torch.tensor(100.0, device=self.sim.device)

        # Set for all
        try:
            self.light.set_falloff(base_val)
        except Exception as e:
            pytest.fail(f"Failed to set falloff for all envs: {e}")

        env_ids = [0, 7, 9]
        new_vals = torch.tensor([200.0, 300.0, 400.0], device=self.sim.device)
        try:
            self.light.set_falloff(new_vals, env_ids=env_ids)
        except Exception as e:
            pytest.fail(f"Failed to set falloff for env_ids={env_ids}: {e}")

    def test_set_and_get_local_pose_matrix_and_vector(self):
        """
        Test setting and getting local pose in both matrix and vector forms.

        1. Set all lights to identity pose (4x4 matrix)
        2. Overwrite subset of lights (env_ids) with custom pose
        3. Check both vector and matrix results match expectations
        """

        # ----------------------------
        # 1. Set all lights to identity matrix
        # ----------------------------
        pose_matrix = torch.eye(4, device=self.sim.device)
        try:
            self.light.set_local_pose(pose_matrix, to_matrix=True)
        except Exception as e:
            pytest.fail(f"Failed to set pose matrix for all envs: {e}")

        result_matrix = self.light.get_local_pose(to_matrix=True)
        assert result_matrix.shape == (
            10,
            4,
            4,
        ), "Unexpected shape from get_local_pose(to_matrix=True)"
        for i, mat in enumerate(result_matrix):
            assert torch.allclose(
                mat, pose_matrix, atol=1e-5
            ), f"Initial matrix pose mismatch at env {i}"

        # ----------------------------
        # 2. Set translation via matrix for selected env_ids
        # ----------------------------
        env_ids = [2, 4, 6]
        pose_matrix_2 = (
            torch.eye(4, device=self.sim.device).unsqueeze(0).repeat(len(env_ids), 1, 1)
        )
        pose_matrix_2[:, 0, 3] = 1.0
        pose_matrix_2[:, 1, 3] = 2.0
        pose_matrix_2[:, 2, 3] = 3.0

        try:
            self.light.set_local_pose(pose_matrix_2, env_ids=env_ids, to_matrix=True)
        except Exception as e:
            pytest.fail(f"Failed to set pose matrix for env_ids={env_ids}: {e}")

        # ----------------------------
        # 3. Check vector form after env_ids modification
        # ----------------------------
        result_vec = self.light.get_local_pose(to_matrix=False)
        assert result_vec.shape == (
            10,
            3,
        ), "Unexpected shape from get_local_pose(to_matrix=False)"

        for i in range(10):
            expected = (
                torch.tensor([1.0, 2.0, 3.0], device=self.sim.device)
                if i in env_ids
                else torch.tensor([0.0, 0.0, 0.0], device=self.sim.device)
            )
            assert torch.allclose(
                result_vec[i], expected, atol=1e-5
            ), f"Translation vector mismatch at env {i}"

        # ----------------------------
        # 4. Verify matrix form translation field
        # ----------------------------
        result_matrix = self.light.get_local_pose(to_matrix=True)
        for i in range(10):
            expected = (
                torch.tensor([1.0, 2.0, 3.0], device=self.sim.device)
                if i in env_ids
                else torch.tensor([0.0, 0.0, 0.0], device=self.sim.device)
            )
            assert torch.allclose(
                result_matrix[i][:3, 3], expected, atol=1e-5
            ), f"Translation matrix mismatch at env {i}"

    def teardown_method(self):
        """Clean up resources after each test method."""
        self.sim.destroy()



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
from enum import Enum
from dataclasses import dataclass
from typing import Callable


class SamplingStrategy(Enum):
    """Sampling strategy for joint space."""

    UNIFORM = "uniform"  # Uniform grid sampling
    RANDOM = "random"  # Random sampling
    HALTON = "halton"  # Quasi-random Halton sequence
    SOBOL = "sobol"  # Quasi-random Sobol sequence
    LATIN_HYPERCUBE = "lhs"  # Latin Hypercube Sampling
    IMPORTANCE = "importance"  # Importance sampling (requires weight function)
    GAUSSIAN = "gaussian"  # Gaussian (normal) distribution sampling
    SPHERE = "sphere"  # Sphere-constrained uniform sampling


@dataclass
class SamplingConfig:
    """Configuration for sampling strategies in workspace analysis."""

    strategy: SamplingStrategy = None
    """Sampling strategy to use."""

    num_samples: int = 1000
    """Number of samples to generate."""

    grid_resolution: int = 10
    """Resolution for grid sampling (used with UNIFORM strategy)."""

    batch_size: int = 1000
    """Number of samples to process in each batch."""

    seed: int = 42
    """Random seed for reproducibility."""

    importance_weight_func: Callable | None = None
    """Weight function for importance sampling (used with IMPORTANCE strategy)."""

    gaussian_mean: float | None = None
    """Mean for Gaussian sampling (used with GAUSSIAN strategy). If None, uses center of bounds."""

    gaussian_std: float | None = None
    """Standard deviation for Gaussian sampling (used with GAUSSIAN strategy). If None, uses 1/6 of range."""

    # Sphere sampling parameters
    sphere_center_mode: str = "bounds_center"
    """How to determine sphere center for SPHERE strategy. Options: 'bounds_center', 'custom', 'auto'."""

    sphere_radius_mode: str = "inscribed"
    """How to determine sphere radius for SPHERE strategy. Options: 'inscribed', 'circumscribed', 'custom'."""

    sphere_boundary_handling: str = "reject"
    """How to handle boundary violations for SPHERE strategy. Options: 'clip', 'reject', 'extend'."""

    sphere_center: list | None = None
    """Custom sphere center for SPHERE strategy (used when sphere_center_mode='custom')."""

    sphere_radius: float | None = None
    """Custom sphere radius for SPHERE strategy (used when sphere_radius_mode='custom')."""

    def __post_init__(self):
        """Set default strategy after initialization."""
        if self.strategy is None:
            self.strategy = SamplingStrategy.UNIFORM



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import torch
import pytest
import numpy as np

from embodichain.lab.sim import SimulationManager, SimulationManagerCfg
from embodichain.lab.sim.objects import Robot
from embodichain.lab.sim.robots import CobotMagicCfg
from embodichain.lab.sim.utility.workspace_analyzer.configs import (
    SamplingConfig,
    VisualizationConfig,
    DimensionConstraint,
)


# Base test class for workspace analyzer
class BaseWorkspaceAnalyzeTest:
    sim = None  # Define as a class attribute

    def setup_simulation(self):
        config = SimulationManagerCfg(headless=True, sim_device="cpu")
        self.sim = SimulationManager(config)
        self.sim.set_manual_update(False)

        cfg_dict = {
            "uid": "CobotMagic",
            "init_pos": [0.0, 0.0, 0.7775],
            "init_qpos": [
                -0.3,
                0.3,
                1.0,
                1.0,
                -1.2,
                -1.2,
                0.0,
                0.0,
                0.6,
                0.6,
                0.0,
                0.0,
                0.05,
                0.05,
                0.05,
                0.05,
            ],
            "solver_cfg": {
                "left_arm": {
                    "class_type": "OPWSolver",
                    "end_link_name": "left_link6",
                    "root_link_name": "left_arm_base",
                    "tcp": [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0.143], [0, 0, 0, 1]],
                },
                "right_arm": {
                    "class_type": "OPWSolver",
                    "end_link_name": "right_link6",
                    "root_link_name": "right_arm_base",
                    "tcp": [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0.143], [0, 0, 0, 1]],
                },
            },
        }

        self.robot: Robot = self.sim.add_robot(cfg=CobotMagicCfg.from_dict(cfg_dict))

    def teardown_method(self):
        """Clean up resources after each test method."""
        self.sim.destroy()


class TestJointWorkspaceAnalyzeTest(BaseWorkspaceAnalyzeTest):
    def setup_method(self):
        self.setup_simulation()

    def test_joint_workspace_analyze(self):
        from embodichain.lab.sim.utility.workspace_analyzer.workspace_analyzer import (
            WorkspaceAnalyzer,
            WorkspaceAnalyzerConfig,
            AnalysisMode,
        )

        joint_config = WorkspaceAnalyzerConfig(
            mode=AnalysisMode.JOINT_SPACE,
            sampling=SamplingConfig(num_samples=100),
            control_part_name="left_arm",
        )
        wa_joint = WorkspaceAnalyzer(
            robot=self.robot,
            config=joint_config,
            sim_manager=self.sim,
        )
        results = wa_joint.analyze(num_samples=100)

        assert "workspace_points" in results
        assert results["workspace_points"].shape[0] > 0
        assert results["mode"] == "joint_space"
        print(
            f"Joint space results: {results['num_valid']}/{results['num_samples']} valid points"
        )


class TestCartesianWorkspaceAnalyzeTest(BaseWorkspaceAnalyzeTest):
    def setup_method(self):
        self.setup_simulation()

    def test_cartesian_workspace_analyze(self):
        from embodichain.lab.sim.utility.workspace_analyzer.workspace_analyzer import (
            WorkspaceAnalyzer,
            WorkspaceAnalyzerConfig,
            AnalysisMode,
        )

        # Get reference pose for IK orientation
        left_qpos = torch.tensor([-0.3, 0.3, 1.0, 1.0, -1.2, -1.2])
        reference_pose = self.robot.compute_fk(
            qpos=left_qpos.unsqueeze(0),
            name="left_arm",
            to_matrix=True,
        )

        cartesian_config = WorkspaceAnalyzerConfig(
            mode=AnalysisMode.CARTESIAN_SPACE,
            sampling=SamplingConfig(
                num_samples=50
            ),  # Smaller sample for faster testing
            constraint=DimensionConstraint(
                min_bounds=[-0.5, -0.5, 0.8],
                max_bounds=[0.5, 0.5, 1.5],
            ),
            reference_pose=reference_pose[0].cpu().numpy(),
            ik_samples_per_point=1,  # Single IK attempt per point for speed
            control_part_name="left_arm",
        )

        wa_cartesian = WorkspaceAnalyzer(
            robot=self.robot,
            config=cartesian_config,
            sim_manager=self.sim,
        )
        results = wa_cartesian.analyze(num_samples=50)

        assert "all_points" in results
        assert "reachable_points" in results
        assert results["all_points"].shape[0] > 0
        assert results["mode"] == "cartesian_space"
        print(
            f"Cartesian space results: {results['num_reachable']}/{results['num_samples']} reachable points"
        )


class TestPlaneWorkspaceAnalyzeTest(BaseWorkspaceAnalyzeTest):
    def setup_method(self):
        self.setup_simulation()

    def test_plane_workspace_analyze(self):
        from embodichain.lab.sim.utility.workspace_analyzer.workspace_analyzer import (
            WorkspaceAnalyzer,
            WorkspaceAnalyzerConfig,
            AnalysisMode,
        )

        # Get reference pose for IK orientation
        left_qpos = torch.tensor([-0.3, 0.3, 1.0, 1.0, -1.2, -1.2])
        reference_pose = self.robot.compute_fk(
            qpos=left_qpos.unsqueeze(0),
            name="left_arm",
            to_matrix=True,
        )

        plane_config = WorkspaceAnalyzerConfig(
            mode=AnalysisMode.PLANE_SAMPLING,
            sampling=SamplingConfig(num_samples=30),  # Small sample for testing
            plane_normal=torch.tensor([0.0, 0.0, 1.0]),  # XY plane
            plane_point=torch.tensor([0.0, 0.0, 1.2]),  # At height 1.2m
            reference_pose=reference_pose[0].cpu().numpy(),
            ik_samples_per_point=1,  # Single IK attempt per point for speed
            visualization=VisualizationConfig(
                enabled=False,  # Disable visualization for testing
            ),
            control_part_name="left_arm",
        )

        wa_plane = WorkspaceAnalyzer(
            robot=self.robot,
            config=plane_config,
            sim_manager=self.sim,
        )
        results = wa_plane.analyze(num_samples=30)

        assert "all_points" in results
        assert "reachable_points" in results
        assert results["all_points"].shape[0] > 0
        assert results["mode"] == "plane_sampling"
        assert "plane_sampling_config" in results
        print(
            f"Plane sampling results: {results['num_reachable']}/{results['num_samples']} reachable points"
        )
        print(
            f"Plane configuration: Normal={results['plane_sampling_config']['plane_normal']}, Point={results['plane_sampling_config']['plane_point']}"
        )


class TestWorkspaceAnalyzerComprehensive(BaseWorkspaceAnalyzeTest):
    """Comprehensive test class for testing consistency between different modes"""

    def setup_method(self):
        self.setup_simulation()

    def test_all_modes_consistency(self):
        """Test basic functionality and result consistency of all three modes"""
        from embodichain.lab.sim.utility.workspace_analyzer.workspace_analyzer import (
            WorkspaceAnalyzer,
            WorkspaceAnalyzerConfig,
            AnalysisMode,
        )

        # Get reference pose
        left_qpos = torch.tensor([-0.3, 0.3, 1.0, 1.0, -1.2, -1.2])
        reference_pose = self.robot.compute_fk(
            qpos=left_qpos.unsqueeze(0),
            name="left_arm",
            to_matrix=True,
        )

        num_samples = 20  # Small sample for fast testing
        results = {}

        # Test joint space mode
        joint_config = WorkspaceAnalyzerConfig(
            mode=AnalysisMode.JOINT_SPACE,
            sampling=SamplingConfig(num_samples=100),
            control_part_name="left_arm",
        )
        wa_joint = WorkspaceAnalyzer(
            robot=self.robot,
            config=joint_config,
            sim_manager=self.sim,
        )
        results["joint"] = wa_joint.analyze(num_samples=num_samples)

        # Test Cartesian space mode
        cartesian_config = WorkspaceAnalyzerConfig(
            mode=AnalysisMode.CARTESIAN_SPACE,
            sampling=SamplingConfig(num_samples=num_samples),
            constraint=DimensionConstraint(
                min_bounds=[-0.5, -0.5, 0.8],
                max_bounds=[0.5, 0.5, 1.5],
            ),
            reference_pose=reference_pose[0].cpu().numpy(),
            ik_samples_per_point=1,
            control_part_name="left_arm",
        )
        wa_cartesian = WorkspaceAnalyzer(
            robot=self.robot,
            config=cartesian_config,
            sim_manager=self.sim,
        )
        results["cartesian"] = wa_cartesian.analyze(num_samples=num_samples)

        # Test plane sampling mode
        plane_config = WorkspaceAnalyzerConfig(
            mode=AnalysisMode.PLANE_SAMPLING,
            sampling=SamplingConfig(num_samples=num_samples),
            plane_normal=torch.tensor([0.0, 0.0, 1.0]),
            plane_point=torch.tensor([0.0, 0.0, 1.2]),
            reference_pose=reference_pose[0].cpu().numpy(),
            ik_samples_per_point=1,
            control_part_name="left_arm",
        )
        wa_plane = WorkspaceAnalyzer(
            robot=self.robot,
            config=plane_config,
            sim_manager=self.sim,
        )
        results["plane"] = wa_plane.analyze(num_samples=num_samples)

        # Verify all modes return valid results
        for mode_name, result in results.items():
            assert "analysis_time" in result
            assert result["analysis_time"] > 0
            assert "mode" in result
            print(
                f"{mode_name.capitalize()} mode: {result['mode']}, Time: {result['analysis_time']:.2f}s"
            )

        # Verify joint space mode
        assert results["joint"]["mode"] == "joint_space"
        assert results["joint"]["num_valid"] > 0

        # Verify Cartesian space mode
        assert results["cartesian"]["mode"] == "cartesian_space"
        assert "reachable_points" in results["cartesian"]

        # Verify plane sampling mode
        assert results["plane"]["mode"] == "plane_sampling"
        assert "plane_sampling_config" in results["plane"]

        print("\n=== Comprehensive Test Summary ===")
        print(
            f"Joint space: {results['joint']['num_valid']}/{results['joint']['num_samples']} valid points"
        )
        print(
            f"Cartesian space: {results['cartesian']['num_reachable']}/{results['cartesian']['num_samples']} reachable points"
        )
        print(
            f"Plane sampling: {results['plane']['num_reachable']}/{results['plane']['num_samples']} reachable points"
        )


if __name__ == "__main__":
    np.set_printoptions(precision=5, suppress=True)
    pytest_args = ["-v", "-s", __file__]
    pytest.main(pytest_args)



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import open3d as o3d
import os

from embodichain.data.dataset import EmbodiChainDataset
from embodichain.data.constants import (
    EMBODICHAIN_DOWNLOAD_PREFIX,
    EMBODICHAIN_DEFAULT_DATA_ROOT,
)

demo_assets = "demo"


class ScoopIceNewEnv(EmbodiChainDataset):
    def __init__(self, data_root: str = None):
        data_descriptor = o3d.data.DataDescriptor(
            os.path.join(
                EMBODICHAIN_DOWNLOAD_PREFIX, demo_assets, "ScoopIceNewEnv.zip"
            ),
            "e92734a9de0f64be33a11fbda0fbd3b6",
        )
        prefix = "ScoopIceNewEnv"
        path = EMBODICHAIN_DEFAULT_DATA_ROOT if data_root is None else data_root

        super().__init__(prefix, data_descriptor, path)


class MultiW1Data(EmbodiChainDataset):
    def __init__(self, data_root: str = None):
        data_descriptor = o3d.data.DataDescriptor(
            os.path.join(EMBODICHAIN_DOWNLOAD_PREFIX, demo_assets, "multi_w1_demo.zip"),
            "984e8fa3aa05cb36a1fd973a475183ed",
        )
        prefix = "MultiW1Data"
        path = EMBODICHAIN_DEFAULT_DATA_ROOT if data_root is None else data_root
        super().__init__(prefix, data_descriptor, path)



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

from langchain_core.messages import SystemMessage
from langchain_core.prompts import (
    ChatPromptTemplate,
    HumanMessagePromptTemplate,
)
from embodichain.utils.utility import encode_image


class CodePrompt:
    @staticmethod
    def one_stage_prompt(**kwargs) -> ChatPromptTemplate:
        prompt = ChatPromptTemplate.from_messages(
            [
                SystemMessage(
                    content="You are an AI assistant that can generate python code to execute robot arms."
                ),
                HumanMessagePromptTemplate.from_template(
                    [
                        {
                            "type": "text",
                            "text": (
                                "Generate a Python code snippet that accomplishes the following task:\n"
                                "{query}\n\n"
                                "You must strictly follow the rules and available functions described below:\n"
                                "{code_prompt}\n\n"
                                "Here are some reference examples of the expected output code:\n"
                                "{code_example}\n\n"
                            ),
                        }
                    ]
                ),
            ]
        )
        return prompt.invoke(kwargs)

    @staticmethod
    def unified_prompt(observations, **kwargs):
        """
        Unified Vision→Code prompt:
        - Model observes the image
        - Understands the scene and the task goal
        - Generates final executable Python code using atomic robot APIs
        """

        # Encode the image
        observation = observations["rgb"]
        kwargs.update({"observation": encode_image(observation)})

        prompt = ChatPromptTemplate.from_messages(
            [
                SystemMessage(
                    content=(
                        "You are a reliable Vision-Language-Code robot assistant. "
                        "You observe an image, understand the scene and the task goal, "
                        "and generate correct Python code using ONLY the allowed atomic robot actions. "
                        "Your final output must be a single Python code block."
                    )
                ),
                HumanMessagePromptTemplate.from_template(
                    [
                        {
                            "type": "image_url",
                            "image_url": {
                                "url": "data:image/png;base64,{observation}",
                            },
                        },
                        {
                            "type": "text",
                            "text": (
                                "### Task Goal\n"
                                "{task_prompt}\n\n"
                                "### Environment Background\n"
                                "{basic_background}\n\n"
                                "### Allowed Atomic Actions\n"
                                "{atom_actions}\n\n"
                                "### Code Rules\n"
                                "{code_prompt}\n\n"
                                "### Reference Code Examples\n"
                                "{code_example}\n\n"
                                "### Final Instructions\n"
                                "Understand the scene from the image and generate final executable Python code "
                                "that performs the task using ONLY the allowed atomic actions.\n\n"
                                "Your entire response must be EXACTLY one Python code block:\n"
                                "```python\n"
                                "# your solution code here\n"
                                "```\n"
                            ),
                        },
                    ]
                ),
            ]
        )

        return prompt.invoke(kwargs)

    @staticmethod
    def one_stage_prompt_according_to_task_plan(**kwargs) -> ChatPromptTemplate:
        prompt = ChatPromptTemplate.from_messages(
            [
                SystemMessage(
                    content=(
                        "You are a reliable robot control code generator.\n"
                        "Your task is to generate Python code that executes robot arm actions.\n\n"
                        "CRITICAL RULES:\n"
                        "- The TASK PLAN defines the available atomic actions, rules, and execution logic.\n"
                        "- You MUST strictly follow the TASK PLAN.\n"
                        "- The CONSTRAINTS section contains additional global constraints you must obey.\n"
                        "- Do NOT invent new actions, functions, parameters, or control flow.\n"
                        "- You MAY include Python comments (# ...) inside the code.\n"
                        "- Your ENTIRE response MUST be a single Python code block.\n"
                        "- The code block MUST be directly executable without modification.\n"
                        "- Do NOT include any text, explanation, or markdown outside the Python code block.\n"
                    )
                ),
                HumanMessagePromptTemplate.from_template(
                    [
                        {
                            "type": "text",
                            "text": (
                                "TASK PLAN (atomic actions, rules, and intended behavior):\n"
                                "{task_plan}\n\n"
                                "GLOBAL CONSTRAINTS (must be satisfied):\n"
                                "{code_prompt}\n\n"
                                "REFERENCE CODE (style and structure only; do NOT copy logic):\n"
                                "{code_example}\n\n"
                                "Generate the corrected Python code now."
                            ),
                        }
                    ]
                ),
            ]
        )
        return prompt.invoke(kwargs)



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import logging

__all__ = ["URDFAssemblyLogger"]


class URDFColorFormatter(logging.Formatter):
    r"""Color formatter for URDF assembly logging"""

    COLORS = {
        "DEBUG": "\033[36m",  # Cyan
        "INFO": "\033[32m",  # Green
        "WARNING": "\033[33m",  # Yellow
        "ERROR": "\033[31m",  # Red
        "CRITICAL": "\033[41m",  # Red background
    }
    # Symbol colors
    BRACKET_COLOR = "\033[94m"  # Bright blue for []
    PAREN_COLOR = "\033[95m"  # Magenta for ()
    RESET = "\033[0m"

    def format(self, record):
        color = self.COLORS.get(record.levelname, self.RESET)
        message = super().format(record)

        # Apply symbol coloring first
        message = self._colorize_symbols(message, color)

        return f"{color}{message}{self.RESET}"

    def _colorize_symbols(self, message, base_color):
        r"""Add colors to brackets and parentheses while preserving base color"""
        import re

        # Color square brackets and their content, then restore base color
        message = re.sub(
            r"\[([^\]]+)\]",
            f"{self.BRACKET_COLOR}[\\1]{self.RESET}{base_color}",
            message,
        )

        # Color parentheses and their content, then restore base color
        message = re.sub(
            r"\(([^)]+)\)", f"{self.PAREN_COLOR}(\\1){self.RESET}{base_color}", message
        )

        return message


class URDFAssemblyLogger:
    r"""URDF Assembly module-specific logger manager"""

    _loggers = {}  # Cache for created loggers
    _initialized = False

    @classmethod
    def get_logger(cls, name: str | None = None) -> logging.Logger:
        r"""Get or create a URDF assembly-specific logger

        Args:
            name: Logger name, defaults to calling module name

        Returns:
            Configured logger instance
        """
        if name is None:
            # Get caller's module name
            import inspect

            frame = inspect.currentframe().f_back
            module_name = frame.f_globals.get("__name__", "unknown")
            if module_name == "__main__":
                name = "urdf_assembly.main"
            else:
                name = f'urdf_assembly.{module_name.split(".")[-1]}'
        else:
            # Ensure using urdf_assembly prefix
            if not name.startswith("urdf_assembly."):
                name = f"urdf_assembly.{name}"

        # Return cached logger or create new one
        if name not in cls._loggers:
            logger = logging.getLogger(name)

            # Avoid duplicate handlers
            if not logger.handlers:
                handler = logging.StreamHandler()
                formatter = URDFColorFormatter("[%(levelname)s] %(name)s: %(message)s")
                handler.setFormatter(formatter)
                logger.addHandler(handler)
                logger.setLevel(logging.INFO)
                logger.propagate = False  # Don't propagate to root logger

            cls._loggers[name] = logger

        return cls._loggers[name]

    @classmethod
    def set_level(cls, level):
        r"""Set log level for all URDF assembly loggers"""
        for logger in cls._loggers.values():
            logger.setLevel(level)

    @classmethod
    def disable_other_loggers(cls):
        r"""Disable output from other non-URDF loggers"""
        logging.getLogger().setLevel(logging.CRITICAL)


# Remove original setup_logger function, replace with URDF-specific initialization
def setup_urdf_logging():
    """Initialize URDF assembly logging system"""
    # Optional: disable other logger outputs
    URDFAssemblyLogger.disable_other_loggers()
    return URDFAssemblyLogger.get_logger("urdf_assembly.main")



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

from embodichain.utils import configclass


@configclass
class AlgorithmCfg:
    """Minimal algorithm configuration shared across RL algorithms."""

    device: str = "cuda"
    learning_rate: float = 3e-4
    batch_size: int = 64
    gamma: float = 0.99
    gae_lambda: float = 0.95
    max_grad_norm: float = 0.5



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import numpy as np
import torch
import ast
from typing import List

from embodichain.utils.logger import log_error, log_warning
from embodichain.lab.gym.utils.misc import mul_linear_expand
from embodichain.lab.sim.planners import (
    MoveType,
    PlanState,
    MotionGenerator,
    MotionGenCfg,
    MotionGenOptions,
    ToppraPlanOptions,
    ToppraPlannerCfg,
)


def draw_axis(env, pose):
    """Draw an axis marker in the simulation for debugging/visualization.

    Args:
        env: The simulation environment.
        pose: The pose (4x4 matrix) where to draw the axis.
    """
    from embodichain.lab.sim.cfg import MarkerCfg

    marker_cfg = MarkerCfg(
        name="test",
        marker_type="axis",
        axis_xpos=pose,
        axis_size=0.01,
        axis_len=0.2,
        arena_index=-1,  # All arenas
    )
    env.sim.draw_marker(cfg=marker_cfg)
    env.sim.update()


def get_arm_states(env, robot_name):
    """Get the current state of the specified robot arm.

    Args:
        env: The simulation environment.
        robot_name: Name of the robot arm (should contain "left" or "right").

    Returns:
        Tuple of (is_left, select_arm, current_qpos, current_pose, current_gripper_state):
            - is_left: bool, whether this is the left arm
            - select_arm: str, arm identifier ("left_arm" or "right_arm")
            - current_qpos: Current joint positions
            - current_pose: Current end-effector pose (4x4 matrix)
            - current_gripper_state: Current gripper state
    """
    left_arm_current_qpos, right_arm_current_qpos = env.get_current_qpos_agent()
    left_arm_current_pose, right_arm_current_pose = env.get_current_xpos_agent()
    left_arm_current_gripper_state, right_arm_current_gripper_state = (
        env.get_current_gripper_state_agent()
    )

    side = "right" if "right" in robot_name else "left"
    is_left = True if side == "left" else False
    select_arm = "left_arm" if is_left else "right_arm"

    arms = {
        "left": (
            left_arm_current_qpos,
            left_arm_current_pose,
            left_arm_current_gripper_state,
        ),
        "right": (
            right_arm_current_qpos,
            right_arm_current_pose,
            right_arm_current_gripper_state,
        ),
    }
    (
        select_arm_current_qpos,
        select_arm_current_pose,
        select_arm_current_gripper_state,
    ) = arms[side]

    return (
        is_left,
        select_arm,
        select_arm_current_qpos,
        select_arm_current_pose,
        select_arm_current_gripper_state,
    )


def find_nearest_valid_pose(env, select_arm, pose, xpos_resolution=0.1):
    """Find the nearest valid pose using reachability validation.

    Args:
        env: The simulation environment.
        select_arm: Arm identifier ("left_arm" or "right_arm").
        pose: Target pose (4x4 matrix).
        xpos_resolution: Resolution for reachability checking.

    Returns:
        torch.Tensor: The nearest valid pose (4x4 matrix).
    """
    # use the validator to choose the nearest valid pose
    # delete the cache every time
    if isinstance(pose, torch.Tensor):
        pose = pose.detach().cpu().numpy()
    ret, _ = env.robot.compute_xpos_reachability(
        select_arm,
        pose,
        xpos_resolution=xpos_resolution,
        qpos_resolution=np.radians(60),
        cache_mode="disk",
        use_cached=False,
        visualize=False,
    )
    ret = np.stack(ret, axis=0)
    # find the nearest valid pose
    xyz = pose[:3, 3]
    ts = np.stack([M[:3, 3] for M in ret], axis=0)  # shape (N,3)
    dists = np.linalg.norm(ts - xyz[None, :], axis=1)
    best_idx = np.argmin(dists)
    nearest_valid_pose = ret[best_idx]
    return torch.from_numpy(nearest_valid_pose)


def get_qpos(env, is_left, select_arm, pose, qpos_seed, force_valid=False, name=""):
    """Solve inverse kinematics to get joint positions for a given pose.

    Args:
        env: The simulation environment.
        is_left: bool, whether this is the left arm.
        select_arm: Arm identifier ("left_arm" or "right_arm").
        pose: Target end-effector pose (4x4 matrix).
        qpos_seed: Seed joint positions for IK solving.
        force_valid: If True, use nearest valid pose if IK fails.
        name: Name for logging purposes.

    Returns:
        Joint positions (qpos) corresponding to the target pose.
    """
    if force_valid:
        try:
            ret, qpos = env.get_arm_ik(pose, is_left=is_left, qpos_seed=qpos_seed)
            if not ret:
                log_error(f"Generate {name} qpos failed.\n")
        except Exception as e:
            log_warning(
                f"Original {name} pose invalid, using nearest valid pose. ({e})\n"
            )
            pose = find_nearest_valid_pose(env, select_arm, pose)

            ret, qpos = env.get_arm_ik(pose, is_left=is_left, qpos_seed=qpos_seed)
    else:
        ret, qpos = env.get_arm_ik(pose, is_left=is_left, qpos_seed=qpos_seed)
        if not ret:
            log_error(f"Generate {name} qpos failed.\n")

    return qpos


def plan_trajectory(
    env,
    select_arm,
    qpos_list,
    sample_num,
    select_arm_current_gripper_state,
    select_qpos_traj,
    ee_state_list_select,
):
    """Plan a trajectory between joint positions and append to trajectory lists.

    Args:
        env: The simulation environment.
        select_arm: Arm identifier ("left_arm" or "right_arm").
        qpos_list: List of joint positions to plan between.
        sample_num: Number of samples for trajectory interpolation.
        select_arm_current_gripper_state: Current gripper state.
        select_qpos_traj: List to append planned joint positions to (modified in-place).
        ee_state_list_select: List to append gripper states to (modified in-place).
    """
    motion_generator = MotionGenerator(
        cfg=MotionGenCfg(planner_cfg=ToppraPlannerCfg(robot_uid=env.robot.uid))
    )

    plan_state = [
        PlanState(qpos=torch.as_tensor(qpos), move_type=MoveType.JOINT_MOVE)
        for qpos in qpos_list
    ]

    ret = motion_generator.generate(
        target_states=plan_state,
        options=MotionGenOptions(
            control_part=select_arm,
            plan_opts=ToppraPlanOptions(
                sample_interval=sample_num,
            ),
        ),
    )

    select_qpos_traj.extend(ret.positions.numpy())
    ee_state_list_select.extend([select_arm_current_gripper_state] * len(ret.positions))


def plan_gripper_trajectory(
    env,
    is_left,
    sample_num,
    execute_open,
    select_arm_current_qpos,
    select_qpos_traj,
    ee_state_list_select,
):
    """Plan a gripper trajectory (opening or closing) and append to trajectory lists.

    Args:
        env: The simulation environment.
        is_left: bool, whether this is the left arm.
        sample_num: Number of samples for gripper motion.
        execute_open: bool, True for opening, False for closing.
        select_arm_current_qpos: Current joint positions.
        select_qpos_traj: List to append joint positions to (modified in-place).
        ee_state_list_select: List to append gripper states to (modified in-place).
    """
    open_state = env.open_state
    close_state = env.close_state

    if execute_open:
        ee_state_expand_select = np.array([close_state, open_state])
        env.set_current_gripper_state_agent(open_state, is_left=is_left)
    else:
        ee_state_expand_select = np.array([open_state, close_state])
        env.set_current_gripper_state_agent(close_state, is_left=is_left)

    ee_state_expand_select = mul_linear_expand(ee_state_expand_select, [sample_num])

    select_qpos_traj.extend([select_arm_current_qpos] * sample_num)
    ee_state_list_select.extend(ee_state_expand_select)


def finalize_actions(select_qpos_traj, ee_state_list_select):
    """Format trajectory data into action format.

    Args:
        select_qpos_traj: List of joint positions.
        ee_state_list_select: List of gripper states.

    Returns:
        np.ndarray: Formatted actions array with joint positions and gripper states.
    """
    # mimic eef state
    actions = np.concatenate(
        [
            np.array(select_qpos_traj),
            np.array(ee_state_list_select),
            np.array(ee_state_list_select),
        ],
        axis=-1,
    )
    return actions


def extract_drive_calls(code_str: str) -> List[str]:
    """Extract all drive() function calls from a code string.

    Args:
        code_str: Python code string to parse.

    Returns:
        List of code blocks containing drive() calls.
    """
    tree = ast.parse(code_str)
    lines = code_str.splitlines()

    drive_blocks = []

    for node in tree.body:
        # Match: drive(...)
        if (
            isinstance(node, ast.Expr)
            and isinstance(node.value, ast.Call)
            and isinstance(node.value.func, ast.Name)
            and node.value.func.id == "drive"
        ):
            # AST line numbers are 1-based
            start = node.lineno - 1
            end = node.end_lineno
            block = "\n".join(lines[start:end])
            drive_blocks.append(block)

    return drive_blocks
